#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from quark.utils import _get_instance_with_sub_options


class Solver(ABC):
    """
    The solver is responsible for finding feasible and high-quality solutions
    of the formulated problem, i.e., of the defined objective function.
    """

    def __init__(self):
        """
        Constructor method.
        """
        self.device_options = []
        self.sub_options = None
        super().__init__()

    @abstractmethod
    def run(self, mapped_problem: any, device: any, config: dict, **kwargs) -> tuple[any, float, dict]:
        """
        This function runs the solving algorithm on a mapped problem instance and returns a solution.

        :param mapped_problem: A representation of the problem that the solver can solve
        :param device: A device the solver can leverage for the algorithm
        :param config: Settings for the solver such as hyperparameters
        :param kwargs: Optional additional settings
        :return: Solution, the time it took to compute it and some optional additional information
        """
        pass

    @abstractmethod
    def get_parameter_options(self) -> dict:
        """
        Method to return the parameters to fine tune the solver.

        Should always be in this format:

        .. code-block:: json

           {
               "parameter_name":{
                  "values":[1, 2, 3],
                  "description":"How many reads do you want?"
               }
           }

        :return: Available solver settings for this solver
        """
        pass

    def get_submodule(self, device_option: str) -> any:
        """
        If self.sub_options is not None, a device is instantiated according to the information given in
        self.sub_options. Otherwise, get_device is called as fall back.

        :param device_option: The option for the device
        :return: Instance of the device class
        """
        if self.sub_options is None:
            return self.get_device(device_option)
        else:
            return _get_instance_with_sub_options(self.sub_options, device_option)

    @abstractmethod
    def get_device(self, device_option: str) -> any:
        """
        Returns the default device based on string. This applies only if
        self.sub_options is None. See get_submodule.

        :param device_option: Desired device option
        :return: Instance of the device class
        """
        pass

    def get_available_device_options(self) -> list:
        """
        Returns the list of available devices.

        :return: List of devices
        """
        if self.sub_options is None:
            return self.device_options
        else:
            return [o["name"] for o in self.sub_options]



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import networkx

from quark.modules.applications.Mapping import Core, Mapping
from quark.utils import start_time_measurement, end_time_measurement


class QIRO(Mapping):
    """
    The quantum-informed recursive optimization (QIRO) formulation for the MIS problem. QIRO recursively simplifies the
    problem classically using information obtained with quantum resources.
    """

    def __init__(self):
        """
        Constructor method
        """
        super().__init__()
        self.submodule_options = ["QrispQIRO"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: list of dict with requirements of this module
        """
        return [{"name": "qrisp", "version": "0.5.2"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return:
        .. code-block:: python

            return {}

        """
        # TODO "optimizer": {
        #          "values": ["not", "yet", "implemented"],
        #          "description": "Which QIRO algorithm should be used?"
        #      }
        return {}

    class Config(TypedDict):
        """
        Attributes of a valid config

        .. code-block:: python
            pass
        """
        pass

    def map(self, problem: networkx.Graph, config: Config) -> tuple[dict, float]:
        """
        Maps the networkx graph to a neutral atom MIS problem.

        :param problem: Networkx graph
        :param config: Config with the parameters specified in Config class
        :return: Dict with neutral MIS, time it took to map it
        """
        start = start_time_measurement()

        qiro_mapped_problem = {
            'graph': problem,
        }
        return qiro_mapped_problem, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "QrispQIRO":
            from quark.modules.solvers.QrispQIRO import QIROSolver  # pylint: disable=C0415
            return QIROSolver()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# copyright Pulp:
# Copyright (c) 2002-2005, Jean-Sebastien Roy
# Modifications Copyright (c) 2007- Stuart Anthony Mitchell
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.

from typing import TypedDict
import pulp

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class MIPaclp(Solver):
    """
    Classical mixed integer problem (MIP) solver for the auto-carrier loading problem (ACLP).
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Local"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "pulp", "version": "2.9.0"},]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: The name of the submodule
        :return: Instance of the default submodule
        """
        if option == "Local":
            from quark.modules.devices.Local import Local  # pylint: disable=C0415
            return Local()
        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns empty dictionary as this solver has no configurable settings.

        :return: Empty dict
        """
        return {}

    class Config(TypedDict):
        """
        Empty config as this solver has no configurable settings.
        """
        pass

    def run(self, mapped_problem: dict, device_wrapper: any, config: Config, **kwargs: dict) \
            -> tuple[dict, float, dict]:
        """
        Solve the ACL problem as a mixed integer problem (MIP).

        :param mapped_problem: Linear problem in form of a dictionary
        :param device_wrapper: Local device
        :param config: Empty dict
        :param kwargs: No additionally settings needed
        :return: Solution, the time it took to compute it and optional additional information
        """
        # Convert dict of problem instance to LP problem
        _, problem_instance = pulp.LpProblem.from_dict(mapped_problem)
        start = start_time_measurement()

        # Solve problem and store relevant data in dictionary
        status = problem_instance.solve()
        obj_value = pulp.value(problem_instance.objective)
        solution_data = {"status": pulp.LpStatus[status], "obj_value": obj_value}
        variables = {}
        for v in problem_instance.variables():
            variables[v.name] = v.varValue
            solution_data["variables"] = variables

        return solution_data, end_time_measurement(start), {}



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from quark.modules.Core import Core


class Circuit(Core, ABC):
    """
    This module is abstract base class for the library-agnostic gate sequence, that define a quantum circuit.
    """

    @abstractmethod
    def generate_gate_sequence(self, input_data: dict, config: any) -> dict:
        """
        Generates the library agnostic gate sequence, a well-defined definition of the quantum circuit.

        :param input_data: Input data required to generate the gate sequence
        :param config: Configuration for the gate sequence
        :return: Generated gate sequence
        """
        pass



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict

import pulser
from pulser.devices import MockDevice
from pulser_simulation import QutipBackend
from pulser.noise_model import NoiseModel
from pulser.backend.config import EmulatorConfig

from quark.modules.devices.pulser.Pulser import Pulser


class MockNeutralAtomDevice(Pulser):
    """
    Class for using the local mock Pulser simulator for neutral atom devices.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__(device_name="mock neutral atom device")
        self.device = MockDevice
        self.backend = QutipBackend
        self.submodule_options = []

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application.

        :return: Configurable settings for the mock neutral atom device
        """
        return {
            "doppler": {
                "values": [False, True],
                "description": "Simulate doppler noise? Has a large impact on performance!"
            },
            "amplitude": {
                "values": [False, True],
                "description": "Simulate amplitude noise? Has a large impact on performance!"
            },
            "SPAM": {
                "values": [False, True],
                "description": "Simulate SPAM noise? Has a large impact on performance!"
            },
            "dephasing": {
                "values": [False, True],
                "description": "Simulate dephasing noise? Has a large impact on performance!"
            },
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.
        """
        doppler: bool
        amplitude: bool
        SPAM: bool
        dephasing: bool

    def get_backend_config(self) -> pulser.backend.config.EmulatorConfig:
        """
        Returns backend configurations.

        :return: Backend config for the emulator
        """
        noise_params = {}
        if "relaxation_rate" in self.config:
            noise_params["relaxation_rate"] = self.config["relaxation_rate"]
        if "dephasing_rate" in self.config:
            noise_params["dephasing_rate"] = self.config["dephasing_rate"]
        if "depolarizing_rate" in self.config:
            noise_params["depolarizing_rate"] = self.config["depolarizing_rate"]

        # Log the noise parameters for debugging
        print("Configured noise parameters:", noise_params)

        # Create the NoiseModel using parameters from self.config
        noise_model = NoiseModel(**noise_params)

        # Return the EmulatorConfig with the noise model
        emulator_config = EmulatorConfig(noise_model=noise_model)
        return emulator_config

    def get_default_submodule(self, option: str) -> None:
        """
        Raises ValueError as this module has no submodules.

        :param option: Option name
        :raises ValueError: If called, since this module has no submodules
        """
        raise ValueError("This module has no submodules.")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union
import logging
from time import perf_counter

import numpy as np

from qiskit.circuit import QuantumCircuit, Parameter
from qiskit.compiler import transpile
from qiskit.transpiler import PassManager, CouplingMap, Layout
from qiskit.transpiler.passes import SetLayout
from qiskit.providers import Backend
from qiskit_aer import Aer, AerSimulator, noise
from qiskit_aer.noise import NoiseModel

from quark.modules.applications.qml.generative_modeling.training.QCBM import QCBM
from quark.modules.applications.qml.generative_modeling.training.Inference import Inference
from quark.modules.applications.qml.generative_modeling.mappings.LibraryGenerative import LibraryGenerative

logging.getLogger("NoisyQiskit").setLevel(logging.WARNING)


def split_string(s):
    return s.split(' ', 1)[0]


class CustomQiskitNoisyBackend(LibraryGenerative):
    """
    This module maps a library-agnostic gate sequence to a qiskit circuit and creates an artificial noise model.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("NoisyQiskit")
        self.submodule_options = ["QCBM", "Inference"]
        self.circuit_transpiled = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "qiskit", "version": "1.3.0"},
            {"name": "qiskit_aer", "version": "0.15.1"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for the Qiskit Library.

        :return: Dictionary with configurable settings
        .. code-block:: python

            return {
                "backend": {
                    "values": ["aer_simulator_gpu", "aer_simulator_cpu"],
                    "description": "Which backend do you want to use? "
                                    "In the NoisyQiskit module only aer_simulators can be used."
                },

                "simulation_method": {
                    "values": ["automatic", "statevector", "density_matrix", "cpu_mps"],  # TODO New names!
                    "description": "What simulation method should be used?"
                },

                "n_shots": {
                    "values": [100, 1000, 10000, 1000000],
                    "description": "How many shots do you want use for estimating the PMF of the model?"
                    # (If 'statevector' was selected as simulation_method, 'n_shots' is only relevant for
                    # studying generalization)"
                },

                "transpile_optimization_level": {
                    "values": [1, 2, 3, 0],
                    "description": "Switch between different optimization levels in the Qiskit transpile"
                    "routine. 1: light optimization, 2: heavy optimization, 3: even heavier optimization,"
                    "0: no optimization. Level 1 recommended as standard option."
                },

                "noise_configuration": {
                    "values": ['Custom configurations', 'No noise'],
                    "description": "What noise configuration do you want to use?"
                },
                "custom_readout_error": {
                    "values": [0, 0.005, 0.01, 0.02, 0.05, 0.07, 0.1, 0.2],
                    "description": "Add a custom readout error."
                },
                "two_qubit_depolarizing_errors": {
                    "values": [0, 0.005, 0.01, 0.02, 0.05, 0.07, 0.1, 0.2],
                    "description": "Add a custom 2-qubit gate depolarizing error."
                },
                "one_qubit_depolarizing_errors": {
                    "values": [0, 0.0001, 0.0005, 0.001, 0.005, 0.007, 0.01, 0.02],
                    "description": "Add a 1-qubit gate depolarizing error."
                },
                "qubit_layout": {
                    # "values": [None, 'linear', 'circle', 'fully_connected', 'ibm_brisbane'],
                    "values": [None, 'linear', 'circle', 'fully_connected'],
                    "description": "How should the qubits be connected in the simulated chip: coupling_map "
                }
            }

        """
        return {
            "backend": {
                "values": ["aer_simulator_gpu", "aer_simulator_cpu"],
                "description": "Which backend do you want to use? "
                               "In the NoisyQiskit module only aer_simulators can be used."
            },

            "simulation_method": {
                "values": ["automatic", "statevector", "density_matrix", "cpu_mps"],  # TODO New names!
                "description": "What simulation method should be used?"
            },

            "n_shots": {
                "values": [100, 1000, 10000, 1000000],
                "description": "How many shots do you want use for estimating the PMF of the model?"
                # (If 'statevector' was selected as simulation_method, 'n_shots' is only relevant for
                # studying generalization)"
            },

            "transpile_optimization_level": {
                "values": [1, 2, 3, 0],
                "description": "Switch between different optimization levels in the Qiskit transpile"
                               "routine. 1: light optimization, 2: heavy optimization, 3: even heavier optimization,"
                               "0: no optimization. Level 1 recommended as standard option."
            },

            "noise_configuration": {
                "values": ['Custom configurations', 'No noise'],
                "description": "What noise configuration do you want to use?"
            },
            "custom_readout_error": {
                "values": [0, 0.005, 0.01, 0.02, 0.05, 0.07, 0.1, 0.2],
                "description": "Add a custom readout error."
            },
            "two_qubit_depolarizing_errors": {
                "values": [0, 0.005, 0.01, 0.02, 0.05, 0.07, 0.1, 0.2],
                "description": "Add a custom 2-qubit gate depolarizing error."
            },
            "one_qubit_depolarizing_errors": {
                "values": [0, 0.0001, 0.0005, 0.001, 0.005, 0.007, 0.01, 0.02],
                "description": "Add a 1-qubit gate depolarizing error."
            },
            "qubit_layout": {
                # "values": [None, 'linear', 'circle', 'fully_connected', 'ibm_brisbane'],
                "values": [None, 'linear', 'circle', 'fully_connected'],
                "description": "How should the qubits be connected in the simulated chip: coupling_map "
            }
        }

    def get_default_submodule(self, option: str) -> Union[QCBM, Inference]:
        """
        Returns the default submodule based on the provided option.

        :param option: The option to select the submodule
        :return: The selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "QCBM":
            return QCBM()
        elif option == "Inference":
            return Inference()
        else:
            raise NotImplementedError(f"Option {option} not implemented")

    def sequence_to_circuit(self, input_data: dict) -> dict:
        """
        Maps the gate sequence, that specifies the architecture of a quantum circuit
        to its Qiskit implementation.

        :param input_data: Collected information of the benchmarking process
        :return: Same dictionary but the gate sequence is replaced by it Qiskit implementation
        """
        n_qubits = input_data["n_qubits"]
        gate_sequence = input_data["gate_sequence"]
        circuit = QuantumCircuit(n_qubits, n_qubits)
        param_counter = 0

        for gate, wires in gate_sequence:
            if gate == "Hadamard":
                circuit.h(wires[0])
            elif gate == "X":
                circuit.x(wires[0])
            elif gate == "SX":
                circuit.sx(wires[0])
            elif gate == "RZ_PI/2":
                circuit.rz(np.pi / 2, wires[0])
            elif gate == "CNOT":
                circuit.cx(wires[0], wires[1])
            elif gate == "ECR":
                circuit.ecr(wires[0], wires[1])
            elif gate == "RZ":
                circuit.rz(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RX":
                circuit.rx(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RY":
                circuit.ry(Parameter(f"x_{param_counter:03d}"), wires[0])
                param_counter += 1
            elif gate == "RXX":
                circuit.rxx(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "RYY":
                circuit.ryy(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "RZZ":
                circuit.rzz(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "CRY":
                circuit.cry(Parameter(f"x_{param_counter:03d}"), wires[0], wires[1])
                param_counter += 1
            elif gate == "Barrier":
                circuit.barrier()
            elif gate == "Measure":
                circuit.measure(wires[0], wires[0])
            else:
                raise NotImplementedError(f"Gate {gate} not implemented")

        input_data["circuit"] = circuit
        input_data.pop("gate_sequence")
        input_data["n_params"] = len(circuit.parameters)

        return input_data

    @staticmethod
    def select_backend(config: str, n_qubits: int) -> Backend:
        """
        This method configures the backend.

        :param config: Name of a backend
        :param n_qubits: Number of qubits
        :return: Configured qiskit backend
        """
        if config == "aer_simulator_gpu":
            backend = Aer.get_backend("aer_simulator")
            backend.set_options(device="GPU")
        elif config == "aer_simulator_cpu":
            backend = Aer.get_backend("aer_simulator")
            backend.set_options(device="CPU")
        else:
            raise NotImplementedError(f"Device Configuration {config} not implemented")

        return backend

    def get_execute_circuit(self, circuit: QuantumCircuit, backend: Backend, config: str, config_dict: dict  # pylint: disable=W0221
                            ) -> tuple[any, any]:
        """
        This method combines the qiskit circuit implementation and the selected backend and returns a function,
        that will be called during training.

        :param circuit: Qiskit implementation of the quantum circuit
        :param backend: Configured qiskit backend
        :param config: Name of a backend
        :param config_dict: Contains information about config
        :return: Tuple that contains a method that executes the quantum circuit for a given set of parameters and the
        transpiled circuit
        """
        n_shots = config_dict["n_shots"]
        n_qubits = circuit.num_qubits
        start = perf_counter()
        backend = self.decompile_noisy_config(config_dict, n_qubits)
        logging.info(f'Backend in Use: {backend=}')
        optimization_level = self.get_transpile_routine(config_dict['transpile_optimization_level'])
        seed_transp = 42  # Remove seed if wanted
        logging.info(f'Using {optimization_level=} with seed: {seed_transp}')
        coupling_map = self.get_coupling_map(config_dict, n_qubits)

        # Create a manual layout if needed (you can customize the layout based on your use case)
        manual_layout = Layout({circuit.qubits[i]: i for i in range(n_qubits)})

        # Define the PassManager
        pass_manager = PassManager()
        pass_manager.append(SetLayout(manual_layout))

        # Run pass manager on the circuit manually (prior to transpiling)
        circuit_passed = pass_manager.run(circuit)

        # Now transpile the circuit after running the pass manager
        circuit_transpiled = transpile(circuit_passed, backend=backend, optimization_level=optimization_level,
                                       seed_transpiler=seed_transp, coupling_map=coupling_map)
        logging.info(f'Circuit operations before transpilation: {circuit.count_ops()}')
        logging.info(f'Circuit operations after transpilation: {circuit_transpiled.count_ops()}')
        logging.info(perf_counter() - start)

        if config in ["aer_simulator_cpu", "aer_simulator_gpu"]:
            def execute_circuit(solutions):

                all_circuits = [circuit_transpiled.assign_parameters(solution) for solution in solutions]
                jobs = backend.run(all_circuits, shots=n_shots)
                samples_dictionary = [jobs.result().get_counts(circuit).int_outcomes() for circuit in all_circuits]
                samples = []
                for result in samples_dictionary:
                    target_iter = np.zeros(2 ** n_qubits)
                    result_keys = list(result.keys())
                    result_vals = list(result.values())
                    target_iter[result_keys] = result_vals
                    target_iter = np.asarray(target_iter)
                    samples.append(target_iter)
                samples = np.asarray(samples)
                pmfs = samples / n_shots

                return pmfs, samples

        else:
            logging.error(f"Unknown backend option selected: {config}")
            logging.error("Run terminates with Exception error.")
            raise Exception

        return execute_circuit, circuit_transpiled

    @staticmethod
    def split_string(s):
        return s.split(' ', 1)[0]

    def decompile_noisy_config(self, config_dict: dict, num_qubits: int) -> Backend:
        """
        This method processes a configuration dictionary.
        If a custom noise configuration is specified, it creates a custom backend configuration; otherwise, it defaults
        to the 'aer_simulator' backend. It returns the configured backend.

        :param config_dict: Contains information about config
        :param num_qubits: Number of qubits
        :return: Configured qiskit backend
        """
        backend_config = config_dict['backend']
        device = 'GPU' if 'gpu' in backend_config else 'CPU'
        simulation_method, device = self.get_simulation_method_and_device(
            device, config_dict['simulation_method']
        )

        backend = self.get_custom_config(config_dict, num_qubits) \
            if config_dict['noise_configuration'] == "Custom configurations" \
            else Aer.get_backend("aer_simulator")

        backend.set_options(device=device, method=simulation_method)
        self.log_backend_options(backend)

        return backend

    def get_simulation_method_and_device(self, device: str, simulation_config: str) -> tuple[str, str]:
        """
        This method specifies the simulation methode and processing unit.

        :param device: Contains information about processing unit
        :param simulation_config: Contains information about qiskit simulation method
        :return: Tuple containing the simulation method and device
        """
        simulation_method = {
            "statevector": "statevector",
            "density_matrix": "density_matrix",
            "cpu_mps": "matrix_product_state"
        }.get(simulation_config, 'automatic')

        if simulation_config == "cpu_mps":
            device = 'CPU'

        return simulation_method, device

    def get_transpile_routine(self, transpile_config: int) -> int:
        """
        This method returns the transpile routine based on the provided configuration.

        :param transpile_config: Configuration for transpile routine
        :return: Transpile routine level
        """
        return transpile_config if transpile_config in [0, 1, 2, 3] else 1

    def get_custom_config(self, config_dict: dict, num_qubits: int) -> Backend:
        """
        This method creates a custom backend configuration based on the provided configuration dictionary.

        :param config_dict: Contains information about config
        :param num_qubits: Number of qubits
        :return: Custom configured qiskit backend
        """
        noise_model = self.build_noise_model(config_dict)
        coupling_map = self.get_coupling_map(config_dict, num_qubits)

        if coupling_map is not None:
            backend = AerSimulator(noise_model=noise_model, coupling_map=coupling_map)
        else:
            backend = AerSimulator(noise_model=noise_model)

        return backend

    def build_noise_model(self, config_dict: dict) -> NoiseModel:
        """
        This method builds a noise model based on the provided configuration dictionary.

        :param config_dict: Contains information about config
        :return: Constructed noise model
        """
        noise_model = noise.NoiseModel()

        if config_dict['custom_readout_error']:
            readout_error = config_dict['custom_readout_error']
            noise_model.add_all_qubit_readout_error(
                [[1 - readout_error, readout_error], [readout_error, 1 - readout_error]]
            )

        self.add_quantum_errors(noise_model, config_dict)
        return noise_model

    def add_quantum_errors(self, noise_model: NoiseModel, config_dict: dict) -> None:
        """
        This method adds quantum errors to the respective operations in the noise model based on the provided
        configuration dictionary.

        :param noise_model: Noise model to which quantum errors are added
        :param config_dict: Contains information about config
        """
        if config_dict['two_qubit_depolarizing_errors'] is not None:
            two_qubit_error = noise.depolarizing_error(config_dict['two_qubit_depolarizing_errors'], 2)
            for gate in ['cx', 'ecr', 'rxx']:
                noise_model.add_all_qubit_quantum_error(two_qubit_error, gate)

        if config_dict['one_qubit_depolarizing_errors'] is not None:
            one_qubit_error = noise.depolarizing_error(config_dict['one_qubit_depolarizing_errors'], 1)
            for gate in ['sx', 'x', 'rx', 'ry', 'rz', 'h', 's']:
                noise_model.add_all_qubit_quantum_error(one_qubit_error, gate)

    def get_coupling_map(self, config_dict: dict, num_qubits: int) -> CouplingMap:
        """
        This method returns the coupling map based on the provided configuration dictionary and number of qubits.

        :param config_dict: Contains information about config
        :param num_qubits: Number of qubits
        :return: Coupling map
        """
        layout = config_dict['qubit_layout']

        if layout == 'linear':
            return CouplingMap.from_line(num_qubits)
        elif layout == 'circle':
            return CouplingMap.from_ring(num_qubits)
        elif layout == 'fully_connected':
            return CouplingMap.from_full(num_qubits)
        elif layout is None:
            logging.info('No coupling map specified, using default.')
            return None
        else:
            raise ValueError(f"Unknown qubit layout: {layout}")

    def log_backend_options(self, backend):
        logging.info(f'Backend configuration: {backend.configuration()}')
        logging.info(f'Simulation method: {backend.options.method}')



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import glob
import argparse
import unittest

from quark.__main__ import handle_benchmark_run, create_benchmark_parser


class TestMain(unittest.TestCase):

    def test_handle_valid_benchmark_run(self) -> None:
        """
        Test a couple of valid QUARK configs
        :return:
        :rtype: None
        """
        parser = argparse.ArgumentParser()
        create_benchmark_parser(parser)

        for filename in glob.glob("tests/configs/valid/**.yml"):
            with self.subTest(msg=f"Running Benchmark Test with valid config {filename}", filename=filename):
                args = parser.parse_args(["--config", filename, "--failfast"])
                self.assertEqual(handle_benchmark_run(args), None)

    def test_handle_invalid_benchmark_run(self) -> None:
        """
        Test a couple of invalid QUARK configs
        :return:
        :rtype: None
        """
        parser = argparse.ArgumentParser()
        create_benchmark_parser(parser)

        for filename in glob.glob("tests/configs/invalid/**.yml"):
            with self.subTest(msg=f"Running Benchmark Test with invalid config {filename}", filename=filename):
                args = parser.parse_args(["--config", filename, "--failfast"])
                with self.assertRaises(Exception):
                    handle_benchmark_run(args)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TypedDict

import numpy as np
from more_itertools import locate
from qiskit_optimization import QuadraticProgram
from qiskit_optimization.converters import QuadraticProgramToQubo

from quark.modules.applications.Mapping import Mapping
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class Ising(Mapping):
    """
    Ising formulation of the auto-carrier loading (ACL) problem.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["QAOA", "QiskitQAOA"]
        self.global_variables = []
        logging.warning("Currently, all scenarios are too large to be solved with an Ising model.")
        logging.warning("Consider using another mapping until the modelling is refined.")

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: list of dict with requirements of this module
        """
        return [
            {"name": "numpy", "version": "1.26.4"},
            {"name": "more-itertools", "version": "10.5.0"},
            {"name": "qiskit-optimization", "version": "0.6.1"},
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns empty dict as this mapping has no configurable settings.

        :return: Empty dict
        """
        return {}

    class Config(TypedDict):
        """
        Empty config as this solver has no configurable settings.
        """
        pass

    def map_pulp_to_qiskit(self, problem: dict) -> QuadraticProgram:
        """
        Maps the problem dict to a quadratic program.

        :param problem: Problem formulation in dict form
        :return: Quadratic program in qiskit-optimization format
        """
        # Details at:
        # https://coin-or.github.io/pulp/guides/how_to_export_models.html
        # https://qiskit.org/documentation/stable/0.26/tutorials/optimization/2_converters_for_quadratic_programs.html

        qp = QuadraticProgram()

        # Variables
        for variable_dict in problem["variables"]:
            if variable_dict["cat"] == "Integer":
                lb = variable_dict["lowBound"]
                ub = variable_dict["upBound"]
                name = variable_dict["name"]
                # If the integer variable is actually a binary variable
                if lb == 0 and ub == 1:
                    qp.binary_var(name)
                # If the integer variable is non-binary
                else:
                    qp.integer_var(lowerbound=lb, upperbound=ub, name=name)

        # Objective function
        obj_arguments = {arg["name"]: arg["value"] for arg in problem["objective"]["coefficients"]}
        # Maximize
        if problem["parameters"]["sense"] == -1:
            qp.maximize(linear=obj_arguments)
        # Minimize
        else:
            qp.minimize(linear=obj_arguments)

        # Constraints
        for constraint in problem["constraints"]:
            const_arguments = {arg["name"]: arg["value"] for arg in constraint["coefficients"]}
            sense = constraint["sense"]
            const_sense = "LE" if sense == -1 else "GE" if sense == 1 else "E"
            qp.linear_constraint(
                linear=const_arguments,
                sense=const_sense,
                rhs=-1 * constraint["constant"],
                name=constraint["name"]
            )

        return qp

    def map(self, problem: dict, config: Config) -> tuple[dict, float]:
        """
        Use Ising mapping of qiskit-optimize.

        :param problem: Dict containing the problem parameters
        :param config: Config with the parameters specified in Config class
        :return: Dict with the Ising, time it took to map it
        """
        start = start_time_measurement()

        # Map Linear problem from dictionary (generated by pulp) to quadratic program
        qp = self.map_pulp_to_qiskit(problem)
        logging.info(qp.export_as_lp_string())

        # convert quadratic problem to qubo to ising
        conv = QuadraticProgramToQubo()
        qubo = conv.convert(qp)

        variables = [variable.name for variable in qubo.variables]
        qubit_op, _ = qubo.to_ising()

        self.global_variables = variables

        # reverse generate J and t out of qubit PauliSumOperator from qiskit
        t_matrix = np.zeros(qubit_op.num_qubits, dtype=complex)
        j_matrix = np.zeros((qubit_op.num_qubits, qubit_op.num_qubits), dtype=complex)

        for pauli_op in qubit_op:
            pauli_str, coeff = pauli_op.to_list()[0]
            index_pos_list = list(locate(pauli_str, lambda a: a == 'Z'))

            if len(index_pos_list) == 1:
                t_matrix[index_pos_list[0]] = coeff
            elif len(index_pos_list) == 2:
                j_matrix[index_pos_list[0]][index_pos_list[1]] = coeff

        return {"J": j_matrix, "t": t_matrix}, end_time_measurement(start)

    def reverse_map(self, solution: dict) -> tuple[dict, float]:
        """
        Maps the solution back to the representation needed by the ACL class for validation/evaluation.

        :param solution: Dict with a bit_string containing the solution
        :return: Solution mapped accordingly, time it took to map it
        """
        start = start_time_measurement()

        if np.any(solution == "-1"):
            solution = self._convert_ising_to_qubo(solution)

        result = {"status": [0]}
        variables = {}
        objective_value = 0

        for bit in solution:
            if solution[bit] > 0:
                if "x" in self.global_variables[bit]:
                    variables[self.global_variables[bit]] = solution[bit]
                    result["status"] = 'Optimal'
                    objective_value += solution[bit]

        result["variables"] = variables
        result["obj_value"] = objective_value

        return result, end_time_measurement(start)

    @staticmethod
    def _convert_ising_to_qubo(solution: any) -> any:
        solution = np.array(solution)
        with np.nditer(solution, op_flags=['readwrite']) as it:
            for x in it:
                if x == -1:
                    x[...] = 0
        return solution

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "QAOA":
            from quark.modules.solvers.QAOA import QAOA  # pylint: disable=C0415
            return QAOA()
        elif option == "QiskitQAOA":
            from quark.modules.solvers.QiskitQAOA import QiskitQAOA  # pylint: disable=C0415
            return QiskitQAOA()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import pickle
import os

from quark.modules.applications.Application import Application
from quark.modules.applications.optimization.Optimization import Optimization
from quark.utils import start_time_measurement, end_time_measurement


class SCP(Optimization):
    """
    The set cover problem (SCP) is a classical combinatorial optimization problem where the objective is to find the
    smallest subset of given elements that covers all required elements in a collection. This can be formulated as
    selecting the minimum number of sets from a collection such that the union of the selected sets contains all
    elements from the universe of the problem instance.

    SCP has widespread applications in various fields, including sensor positioning, resource allocation, and network
    design. For example, in sensor positioning, SCP can help determine the fewest number of sensors required to cover
    a given area. Similarly, in resource allocation, SCP helps to allocate resources in an optimal way, ensuring
    coverage of all demand points while minimizing costs. Network design also uses SCP principles to efficiently place
    routers or gateways in a network to ensure full coverage with minimal redundancy.

    This implementation of SCP provides configurable problem instances of different sizes, such as "Tiny," "Small,"
    and "Large," allowing the user to explore solutions with varying complexities. We employ various quantum-inspired
    methods to solve SCP, including a mapping to the QUBO (Quadratic Unconstrained Binary Optimization) formulation
    using Qubovert. These approaches allow us to explore how different optimization algorithms and frameworks perform
    when applied to this challenging problem, offering insights into both classical and emerging quantum methods.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("SCP")
        self.submodule_options = ["qubovertQUBO"]

    def get_solution_quality_unit(self) -> str:
        return "Number of selected subsets"

    def get_default_submodule(self, option: str) -> Application:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "qubovertQUBO":
            from quark.modules.applications.optimization.SCP.mappings.qubovertQUBO import QubovertQUBO  # pylint: disable=C0415
            return QubovertQUBO()
        else:
            raise NotImplementedError(f"Mapping Option {option} not implemented")

    def get_parameter_options(self):
        """
        Returns the configurable settings for this application

        :return: Dictionary containing parameter options
        .. code-block:: python

        return {
            "model_select": {
                "values": list(["Tiny", "Small", "Large"]),
                "description": "Please select the problem size(s). Tiny: 4 elements, 3 subsets. Small:
                15 elements, 8 subsets. Large: 100 elements, 100 subsets"
            }
        }
        """
        return {
            "model_select": {
                "values": list(["Tiny", "Small", "Large"]),
                "description": "Please select the problem size(s). Tiny: 4 elements, 3 subsets. Small: 15 elements, "
                               "8 subsets. Large: 100 elements, 100 subsets"
            }
        }

    class Config(TypedDict):
        model_select: str

    def generate_problem(self, config: Config) -> tuple[set, list]:
        """
        Generates predefined instances of the SCP.

        :param config: Config specifying the selected problem instances
        :return: The union of all elements of an instance and a set of subsets, each covering a part of the union
        """
        model_select = config['model_select']
        self.application = {}

        if model_select == "Tiny":
            self.application["elements_to_cover"] = set(range(1, 4))
            self.application["subsets"] = [{1, 2}, {1, 3}, {3, 4}]
        elif model_select == "Small":
            self.application["elements_to_cover"] = set(range(1, 15))
            self.application["subsets"] = [
                {1, 3, 4, 6, 7, 13}, {4, 6, 8, 12}, {2, 5, 9, 11, 13}, {1, 2, 7, 14, 15},
                {3, 10, 12, 14}, {7, 8, 14, 15}, {1, 2, 6, 11}, {1, 2, 4, 6, 8, 12}
            ]

        elif model_select == "Large":
            self.application["elements_to_cover"] = set(range(1, 100))
            self.application["subsets"] = []
            path = os.path.join(os.path.dirname(__file__))
            with open(f"{path}/data/set_cover_data_large.txt") as data:
                while line := data.readline():
                    new_set = []
                    for i in line.split(','):
                        new_set.append(int(i))
                    new_set = set(new_set)
                    self.application["subsets"].append(new_set)

        else:
            raise ValueError(f"Unknown model_select value: {model_select}")

        return self.application["elements_to_cover"], self.application["subsets"]

    def process_solution(self, solution: list) -> tuple[list, float]:
        """
        Returns list of selected subsets and the time it took to process the solution.

        :param solution: Unprocessed solution
        :return: Processed solution and the time it took to process it
        """
        start_time = start_time_measurement()
        selected_subsets = [list(self.application["subsets"][i]) for i in solution]
        return selected_subsets, end_time_measurement(start_time)

    def validate(self, solution: list) -> tuple[bool, float]:
        """
        Checks if the elements of the subsets that are part of the solution cover every element of the instance.

        :param solution: List containing all subsets that are part of the solution
        :return: Boolean whether the solution is valid and time it took to validate
        """
        start = start_time_measurement()
        covered = set.union(*[set(subset) for subset in solution])

        return covered == self.application["elements_to_cover"], end_time_measurement(start)

    def evaluate(self, solution: list) -> tuple[int, float]:
        """
        Calculates the number of subsets that are of the solution.

        :param solution: List containing all subsets that are part of the solution
        :return: Number of subsets and the time it took to calculate it
        """
        start = start_time_measurement()
        selected_num = len(solution)

        return selected_num, end_time_measurement(start)

    def save(self, path: str, iter_count: int) -> None:
        """
        Saves the SCP instance to a file.

        :param path: Path to save the SCP instance
        :param iter_count: Iteration count
        """
        with open(f"{path}/SCP_instance", "wb") as file:
            pickle.dump(self.application, file, pickle.HIGHEST_PROTOCOL)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import logging

import numpy as np
import pulser

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class NeutralAtomMIS(Solver):
    """
    Neutral atom quantum computer maximum independent sets solver.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["MockNeutralAtomDevice"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "pulser", "version": "1.1.1"}]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: The name of the submodule
        :return: Instance of the default submodule
        """
        if option == "MockNeutralAtomDevice":
            from quark.modules.devices.pulser.MockNeutralAtomDevice import MockNeutralAtomDevice  # pylint: disable=C0415
            return MockNeutralAtomDevice()
        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this solver.

        :return: Dictionary of configurable settings.
        """
        return {
            "samples": {
                "values": [10, 100, 1000, 10000],
                "custom_input": True,
                "allow_ranges": True,
                "postproc": int,
                "description": "How many samples from the quantum computer do you want per measurement?"
            },
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        samples (int): How many times to sample the final state from the quantum computer per measurement
        """
        samples: int

    def run(self, mapped_problem: dict, device_wrapper: any, config: any, **kwargs: dict) -> tuple[list, float, dict]:
        """
        The given application is a problem instance from the pysat library. This uses the rc2 maxsat solver
        given in that library to return a solution.

        :param mapped_problem: Dictionary with graph and register
        :param device_wrapper: Device to run the problem on
        :param config: Solver Configuration
        :param kwargs: Additional settings (not used)
        :return: Solution, the time it took to compute it and optional additional information
        """
        register = mapped_problem.get('register')
        graph = mapped_problem.get('graph')
        nodes = list(graph.nodes())
        edges = list(graph.edges())

        logging.info(f"Got problem with {len(graph.nodes)} nodes, {len(graph.edges)} edges.")

        device = device_wrapper.get_device()
        device.validate_register(register)
        device_backend = device_wrapper.get_backend()
        device_config = device_wrapper.get_backend_config()

        start = start_time_measurement()

        sequence = self._create_sequence(register, device)

        if device_wrapper.device_name == "mock neutral atom device":
            backend = device_backend(sequence, device_config)
            results = backend.run(progress_bar=False)
            sampled_state_counts = results.sample_final_state(N_samples=config['samples'])
        else:
            raise NotImplementedError(f"Device Option {device_wrapper.device_name} not implemented")

        valid_state_counts = self._filter_invalid_states(sampled_state_counts, nodes, edges)
        state = self._select_best_state(valid_state_counts, nodes)
        state_nodes = self._translate_state_to_nodes(state, nodes)

        return state_nodes, end_time_measurement(start), {}

    def _create_sequence(self, register: pulser.Register, device: pulser.devices._device_datacls.Device) \
            -> pulser.Sequence:
        """
        Creates a pulser sequence from a register and a device.

        :param register: The quantum register
        :param device: The device being used
        :return: The created sequence
        """
        pulses = self._create_pulses(device)
        sequence = pulser.Sequence(register, device)
        sequence.declare_channel("Rydberg global", "rydberg_global")
        for pulse in pulses:
            sequence.add(pulse, "Rydberg global")
        return sequence

    def _create_pulses(self, device: pulser.devices._device_datacls.Device) -> list[pulser.Pulse]:
        """
        Creates pulses tuned to MIS problem.

        Pulse creation is a whole art/science on its own that we have not delved into yet.
        If you shape and finetune your pulses to decrease compute time on your neutral atom device.
        We found this configuration in the documentation of the pulser documentation and it works for MIS.
        We are hesitant to make them parametrizable, because setting the wrong values will break your whole MIS.
        Though parameterization of pulses is a feature that we might implement in the future.

        :param device: The device being used
        :return: List of pulses
        """
        omega_max = 2.3 * 2 * np.pi
        delta_factor = 2 * np.pi

        channel = device.channels['rydberg_global']
        max_amp = channel.max_amp
        if max_amp is not None and max_amp < omega_max:
            omega_max = max_amp

        delta_0 = -3 * delta_factor
        delta_f = 1 * delta_factor

        t_rise = 2000
        t_fall = 2000
        t_sweep = (delta_f - delta_0) / (2 * np.pi * 10) * 5000

        rise = pulser.Pulse.ConstantDetuning(
            pulser.waveforms.RampWaveform(t_rise, 0.0, omega_max), delta_0, 0.0
        )
        sweep = pulser.Pulse.ConstantAmplitude(
            omega_max, pulser.waveforms.RampWaveform(t_sweep, delta_0, delta_f), 0.0
        )
        fall = pulser.Pulse.ConstantDetuning(
            pulser.waveforms.RampWaveform(t_fall, omega_max, 0.0), delta_f, 0.0
        )
        pulses = [rise, sweep, fall]

        for pulse in pulses:
            channel.validate_pulse(pulse)

        return pulses

    def _filter_invalid_states(self, state_counts: dict, nodes: list, edges: list) -> dict:
        """
        Filters out invalid states that do not meet the problem constraints.

        :param state_counts: Counts of each sampled data
        :param nodes: List of nodes in the graph
        :param edges: List of edges in the graph
        :return: Dictionary of valid state counts
        """
        valid_state_counts = {}
        for state, count in state_counts.items():
            selected_nodes = self._translate_state_to_nodes(state, nodes)

            is_valid = True
            for edge in edges:
                if edge[0] in selected_nodes and edge[1] in selected_nodes:
                    is_valid = False
                    break
            if is_valid:
                valid_state_counts[state] = count

        return valid_state_counts

    def _translate_state_to_nodes(self, state: str, nodes: list) -> list:
        """
        Translates a state string into the corresponding list of nodes.

        :param state: State string
        :param nodes: List of nodes
        :return: List of nodes corresponding to the states
        """
        return [key for index, key in enumerate(nodes) if state[index] == '1']

    def _select_best_state(self, states: dict, nodes: list) -> str:
        """
        Selects the best state from the available valid states.

        :param states: Dictionary of valid states and their counts
        :param nodes: List of nodes
        :return: The best state as a string
        """
        # TODO: Implement the samplers
        try:
            best_state = max(states, key=lambda k: states[k])
        except Exception:  # pylint: disable=W0702
            # TODO: Specify error
            # TODO: Clean this up
            n_nodes = len(nodes)
            best_state = "0" * n_nodes

        return best_state



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union
import numpy as np

from quark.modules.applications.qml.generative_modeling.transformations.Transformation import Transformation
from quark.modules.applications.qml.generative_modeling.circuits.CircuitStandard import CircuitStandard
from quark.modules.applications.qml.generative_modeling.circuits.CircuitCardinality import CircuitCardinality


class MinMax(Transformation):  # pylint: disable=R0902
    """
    In min-max normalization each data point is shifted
    such that it lies between 0 and 1.
    """

    def __init__(self):
        super().__init__("MinMax")
        self.submodule_options = ["CircuitStandard", "CircuitCardinality"]
        self.transform_config = None
        self.max = None
        self.min = None
        self.n_qubits = None
        self.dataset = None
        self.dataset_name = None
        self.grid_shape = None
        self.histogram_train = None
        self.histogram_train_original = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "numpy", "version": "1.26.4"}]

    def get_default_submodule(self, option: str) -> Union[CircuitStandard, CircuitCardinality]:
        if option == "CircuitStandard":
            return CircuitStandard()
        elif option == "CircuitCardinality":
            return CircuitCardinality()
        else:
            raise NotImplementedError(f"Circuit Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns empty dict as this transformation has no configurable settings.

        :return: Empty dict
        """
        return {}

    def transform(self, input_data: dict, config: dict) -> dict:
        """
        Transforms the input dataset using MinMax transformation and computes histograms
        of the training dataset in the transformed space.

        :param input_data: A dictionary containing information about the dataset and application configuration.
        :param config: A dictionary with parameters specified in the Config class.
        :return: A tuple containing a dictionary with MinMax-transformed data.
        """
        self.dataset_name = input_data["dataset_name"]
        self.dataset = input_data["dataset"]
        self.n_qubits = input_data["n_qubits"]
        self.grid_shape = int(2 ** (self.n_qubits // 2))
        n_registers = self.dataset.shape[-1]

        # Calculate ranges for the original dataset and the transformed dataset
        ranges_original = np.column_stack((np.min(self.dataset, axis=0), np.max(self.dataset, axis=0)))
        transformed_dataset = self.fit_transform(self.dataset)
        ranges_transformed = np.column_stack((np.min(transformed_dataset, axis=0), np.max(transformed_dataset, axis=0)))

        # Compute histogram for the transformed dataset
        transformed_histogram_grid = np.histogramdd(
            transformed_dataset,
            bins=self.grid_shape,
            range=ranges_transformed)[0]
        histogram_transformed_1d = transformed_histogram_grid.flatten()

        solution_space = np.zeros(len(transformed_dataset), dtype=int)
        # Initialize a variable to keep track of the current position in the result_array
        position = 0
        value = 0
        for count in histogram_transformed_1d:
            if count > 0:
                solution_space[position:position + int(count)] = value
                position += int(count)
            value += 1

        binary_strings = [np.binary_repr(x, width=self.n_qubits) for x in solution_space]
        binary_transformed = np.array([list(map(int, s)) for s in binary_strings])
        learned_histogram = np.histogramdd(self.dataset, bins=self.grid_shape, range=ranges_original)
        self.histogram_train_original = learned_histogram[0] / np.sum(learned_histogram[0])

        # Compute histogram for the transformed dataset
        learned_histogram = np.histogramdd(transformed_dataset, bins=self.grid_shape, range=ranges_transformed)
        histogram_train = learned_histogram[0] / np.sum(learned_histogram[0])
        self.histogram_train = histogram_train.flatten()

        self.transform_config = {
            "histogram_train": self.histogram_train,
            "binary_train": binary_transformed,
            "dataset_name": self.dataset_name,
            "n_registers": n_registers,
            "n_qubits": self.n_qubits,
            "train_size": input_data["train_size"],
            "store_dir_iter": input_data["store_dir_iter"]
        }

        return self.transform_config

    def reverse_transform(self, input_data: dict) -> dict:
        """
        Transforms the solution back to the representation needed for validation/evaluation.

        :param input_data: Dictionary containing the solution
        :return: Solution transformed accordingly
        """
        best_results = input_data["best_sample"]
        depth = input_data["depth"]
        architecture_name = input_data["architecture_name"]
        n_qubits = input_data["n_qubits"]
        n_registers = self.transform_config["n_registers"]
        KL_best_transformed = min(input_data["KL"])
        circuit_transpiled = input_data['circuit_transpiled']

        array_bins = Transformation.compute_discretization_efficient(n_qubits, n_registers)
        transformed_samples = Transformation.generate_samples_efficient(best_results, array_bins, n_registers,
                                                                        noisy=True)

        # Calculate ranges for the transformed samples
        ranges_transformed = np.column_stack((np.min(transformed_samples, axis=0), np.max(transformed_samples, axis=0)))

        # Compute histogram for the transformed samples
        learned_histogram = np.histogramdd(transformed_samples, bins=self.grid_shape, range=ranges_transformed)
        histogram_generated_transformed = learned_histogram[0] / np.sum(learned_histogram[0])
        histogram_generated_transformed = histogram_generated_transformed.flatten()

        original_samples = self.inverse_transform(transformed_samples)

        # Calculate ranges for the original samples
        ranges_original = np.column_stack((np.min(original_samples, axis=0), np.max(original_samples, axis=0)))

        # Compute histogram for the original samples
        learned_histogram = np.histogramdd(original_samples, bins=self.grid_shape, range=ranges_original)
        histogram_generated_original = learned_histogram[0] / np.sum(learned_histogram[0])
        histogram_generated_original = histogram_generated_original.flatten()

        best_parameter = input_data["best_parameter"]

        reverse_config_trans = {
            "generated_samples": best_results,
            "transformed_samples": transformed_samples,
            "depth": depth,
            "architecture_name": architecture_name,
            "dataset_name": self.dataset_name,
            "n_qubits": n_qubits,
            "best_parameter": best_parameter,
            "histogram_train": self.histogram_train,
            "histogram_train_original": self.histogram_train_original,
            "histogram_generated_original": histogram_generated_original,
            "histogram_generated": histogram_generated_transformed,
            "KL_best_transformed": KL_best_transformed,
            "store_dir_iter": input_data["store_dir_iter"],
            "circuit_transpiled": circuit_transpiled
        }

        return reverse_config_trans

    def fit_transform(self, data: np.ndarray) -> np.ndarray:
        """
        Method that performs the min max normalization.

        :param data: Data to be fitted
        :return: Fitted data
        """
        data_min = data.min()
        data_max = data.max() - data_min
        data = (data - data_min) / data_max

        return data

    def inverse_transform(self, data: np.ndarray) -> np.ndarray:
        """
        Method that performs the inverse min max normalization.

        :param data: Data to be fitted
        :return: Data in original space
        """
        data_min = data.min()
        data_max = data.max() - data_min

        return data * data_max + data_min



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
from typing import TypedDict
import pickle
import logging
import os

import networkx as nx
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
import numpy as np

from quark.modules.applications.Application import Core
from quark.modules.applications.optimization.Optimization import Optimization
from quark.utils import start_time_measurement, end_time_measurement


class PVC(Optimization):
    """
    In modern vehicle manufacturing, robots take on a significant workload, including performing welding
    jobs, sealing welding joints, or applying paint to the car body. While the robot’s tasks vary widely,
    the objective remains the same: Perform a job with the highest possible quality in the shortest amount
    of time, optimizing efficiency and productivity on the manufacturing line.

    For instance, to protect a car’s underbody from corrosion, exposed welding seams are sealed
    by applying a polyvinyl chloride layer (PVC). The welding seams need to be traversed by a robot to
    apply the material. It is related to TSP, but different and even more complex in some aspects.

    The problem of determining the optimal route for robots to traverse all seams shares similarities
    with Traveling Salesman Problem (TSP), as it involves finding the shortest possible route to
    visit multiple locations. However, it introduces additional complexities, such as different tool
    and configuration requirements for each seam, making it an even more challenging problem to solve.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("PVC")
        self.submodule_options = [
            "Ising", "QUBO", "GreedyClassicalPVC", "ReverseGreedyClassicalPVC", "RandomPVC"
        ]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "networkx", "version": "3.4.2"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_solution_quality_unit(self) -> str:
        """
        Returns the unit of measure for solution quality.

        :return: Unit of measure for solution quality
        """
        return "Tour cost"

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "Ising":
            from quark.modules.applications.optimization.PVC.mappings.ISING import Ising  # pylint: disable=C0415
            return Ising()
        elif option == "QUBO":
            from quark.modules.applications.optimization.PVC.mappings.QUBO import QUBO  # pylint: disable=C0415
            return QUBO()
        elif option == "GreedyClassicalPVC":
            from quark.modules.solvers.GreedyClassicalPVC import GreedyClassicalPVC  # pylint: disable=C0415
            return GreedyClassicalPVC()
        elif option == "ReverseGreedyClassicalPVC":
            from quark.modules.solvers.ReverseGreedyClassicalPVC import ReverseGreedyClassicalPVC  # pylint: disable=C0415
            return ReverseGreedyClassicalPVC()
        elif option == "RandomPVC":
            from quark.modules.solvers.RandomClassicalPVC import RandomPVC  # pylint: disable=C0415
            return RandomPVC()
        else:
            raise NotImplementedError(f"Mapping Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application.

        :return: Dictionary containing parameter options
        .. code-block:: python

            return {
                    "seams": {
                        "values": list(range(1, 18)),
                        "description": "How many seams should be sealed?"
                    }
                }
        """
        return {
            "seams": {
                "values": list(range(1, 18)),
                # In the current implementation the graph can only be as large as the reference input graph
                "description": "How many seams should be sealed?"
            }
        }

    class Config(TypedDict):
        """
       Configuration attributes for PVC problem generation.

       Attributes:
        seams (int): Number of seams for the graph
        """
        seams: int

    def generate_problem(self, config: Config) -> nx.Graph:
        """
        Uses the reference graph to generate a problem for a given config.

        :param config: Config specifying the number of seams for the problem
        :return: Networkx graph representing the problem
        """
        if config is None:
            config = {"seams": 3}
        seams = config['seams']

        # Read in the original graph
        with open(os.path.join(os.path.dirname(__file__), "data", "reference_graph.gpickle"), "rb") as file:
            graph = pickle.load(file)

        # Get number of seam in graph
        seams_in_graph = list({x[0] for x in graph.nodes})
        seams_in_graph.sort()
        seams_in_graph.remove(0)  # Always need the base node 0 (which is not a seam)

        if len(seams_in_graph) < seams:
            logging.info("Too many seams! The original graph has less seams than that!")

        unwanted_seams = seams_in_graph[-len(seams_in_graph) + seams:]
        unwanted_nodes = [x for x in graph.nodes if x[0] in unwanted_seams]

        for node in unwanted_nodes:
            graph.remove_node(node)

        if not nx.is_strongly_connected(graph):
            logging.error("Graph is not connected!")
            raise ValueError("Graph is not connected!")

        # Gather unique configurations and tools
        config = [x[2]['c_start'] for x in graph.edges(data=True)]
        config = list(set(config + [x[2]['c_end'] for x in graph.edges(data=True)]))
        tool = [x[2]['t_start'] for x in graph.edges(data=True)]
        tool = list(set(tool + [x[2]['t_end'] for x in graph.edges(data=True)]))

        # Fill the rest of the missing edges with high values
        current_edges = [
            (edge[0], edge[1], edge[2]['t_start'], edge[2]['t_end'], edge[2]['c_start'], edge[2]['c_end'])
            for edge in graph.edges(data=True)
        ]
        all_possible_edges = list(itertools.product(list(graph.nodes), repeat=2))
        all_possible_edges = [
            (edges[0], edges[1], t_start, t_end, c_start, c_end)
            for edges in all_possible_edges
            for c_end in config
            for c_start in config
            for t_end in tool
            for t_start in tool if edges[0] != edges[1]
        ]

        missing_edges = [item for item in all_possible_edges if item not in current_edges]

        # Add these edges with very high values
        for edge in missing_edges:
            graph.add_edge(
                edge[0], edge[1], c_start=edge[4], t_start=edge[2], c_end=edge[5], t_end=edge[3], weight=100000
            )

        logging.info("Created PVC problem with the following attributes:")
        logging.info(f" - Number of seams: {seams}")
        logging.info(f" - Number of different configs: {len(config)}")
        logging.info(f" - Number of different tools: {len(tool)}")

        self.application = graph
        return graph.copy()

    def process_solution(self, solution: dict) -> tuple[list, float]:
        """
        Converts solution dictionary to list of visited seams.

        :param solution: Unprocessed solution
        :return: Processed solution and the time it took to process it
        """
        start_time = start_time_measurement()
        nodes = list(self.application.nodes())
        start = ((0, 0), 1, 1)
        route: list = [None] * int((len(self.application) - 1) / 2 + 1)
        visited_seams = []

        if sum(value == 1 for value in solution.values()) > len(route):
            logging.warning("Result is longer than route! This might be problematic!")

        # Prevent duplicate node entries by enforcing only one occurrence per node along route
        for (node, config, tool, timestep), val in solution.items():
            if val and (node[0] not in visited_seams):
                if route[timestep] is not None:
                    visited_seams.remove(route[timestep][0][0])
                route[timestep] = (node, config, tool)
                visited_seams.append(node[0])

        # Fill missing values in the route
        if None in route:
            logging.info(f"Route until now is: {route}")
            nodes_unassigned = [(node, 1, 1) for node in nodes if node[0] not in visited_seams]
            nodes_unassigned = list(np.random.permutation(nodes_unassigned, dtype=object))
            logging.info(nodes_unassigned)
            logging.info(visited_seams)
            logging.info(nodes)
            for idx, node in enumerate(route):
                if node is None:
                    route[idx] = nodes_unassigned.pop(0)

        # Cycle solution to start at provided start location
        if start is not None and route[0] != start:
            idx = route.index(start)
            route = route[idx:] + route[:idx]

        parsed_route = ' ->\n'.join(
            [
                f' Node {visit[0][1]} of Seam {visit[0][0]} using config '
                f' {visit[1]} & tool {visit[2]}'
                for visit in route
            ]
        )
        logging.info(f"Route found:\n{parsed_route}")

        return route, end_time_measurement(start_time)

    def validate(self, solution: list) -> tuple[bool, float]:
        """
        Checks if all seams and the home position are visited for a given solution.

        :param solution: List containing the nodes of the solution
        :return: Boolean whether the solution is valid and time it took to validate
        """
        # Check if all seams are visited in route
        start = start_time_measurement()
        visited_seams = {seam[0][0] for seam in solution if seam is not None}

        if len(visited_seams) == len(solution):
            logging.info(f"All {len(solution) - 1} seams and "
                         "the base node got visited (We only need to visit 1 node per seam)")
            return True, end_time_measurement(start)
        else:
            logging.error(f"Only {len(visited_seams) - 1} got visited")
            return False, end_time_measurement(start)

    def evaluate(self, solution: list) -> tuple[float, float]:
        """
        Calculates the tour length for a given valid tour.

        :param solution: List containing the nodes of the solution
        :return: Tour length, time it took to calculate the tour length
        """
        start = start_time_measurement()

        # Get the total distance
        total_dist = 0
        for idx, _ in enumerate(solution[:-1]):
            edge = next(
                item for item in list(self.application[solution[idx][0]][solution[idx + 1][0]].values())
                if item["c_start"] == solution[idx][1] and item["t_start"] == solution[idx][2] and
                item["c_end"] == solution[idx + 1][1] and item["t_end"] == solution[idx + 1][2]
            )
            dist = edge['weight']
            total_dist += dist
        logging.info(f"Total distance (without return): {total_dist}")

        # Add distance between start and end point to complete cycle
        return_edge = next(
            item for item in list(self.application[solution[0][0]][solution[-1][0]].values())
            if item["c_start"] == solution[0][1] and item["t_start"] == solution[0][2] and
            item["c_end"] == solution[-1][1] and item["t_end"] == solution[-1][2]
        )
        return_distance = return_edge['weight']
        logging.info(f"Distance between start and end: {return_distance}")

        # Get distance for full cycle
        distance = total_dist + return_distance
        logging.info(f"Total distance (including return): {distance}")

        return distance, end_time_measurement(start)

    def save(self, path: str, iter_count: int) -> None:
        """
        Saves the generated problem graph to a file.

        :param path: Path to save the problem graph
        :param iter_count: Iteration count for file versioning
        """
        with open(f"{path}/graph_iter_{iter_count}.gpickle", "wb") as file:
            pickle.dump(self.application, file, pickle.HIGHEST_PROTOCOL)

    def visualize_solution(self, processed_solution, path: str):
        """
        Plot a graph representing the possible locations where seams can start or end, with arrows representing either idle movements or the sealing of a seam

        :param processed_solution: The solution already processed by :func:`process_solution`, a list of tuples representing seam start points and the config and tool needed to seal the seam.
        :param path: File path for the plot
        :returns: None
        """
        NODE_SIZE = 300   # Default=300
        EDGE_WIDTH = 1.0  # Default=1.0
        FONT_SIZE = 12    # Default=12

        highest_node_id = max(node[1] for node in self.application.nodes())
        G = nx.MultiDiGraph()
        G.add_nodes_from(range(highest_node_id + 1))
        pos = nx.circular_layout(G)

        tools = set()
        configs = set()
        current_node = 0
        for ((seam1, node1), config, tool) in processed_solution[1:]:
            config = config - 1
            tools.add(tool)
            configs.add(config)
            (seam2, node2) = next((seam, node)
                                  for (seam, node) in self.application.nodes() if seam == seam1 and not node == node1)
            assert seam1 == seam2, "This is bad"
            if not current_node == node1:
                G.add_edge(current_node, node1, color=7, width=EDGE_WIDTH, style=-1)
            G.add_edge(node1, node2, color=tool, width=2 * EDGE_WIDTH, style=config)
            current_node = node2

        # The 8 here controls how many edges between the same two nodes are at
        # most drawn with spacing between them before drawing them on top of each
        # other to avoid cluttering
        connectionstyle = [f"arc3,rad={r}" for r in itertools.accumulate([0.15] * 8)]
        style_options = ["solid", "dotted", "dashed", "dashdot"]
        cmap = plt.cm.Dark2
        tools = list(tools)
        configs = list(configs)
        legend_elements = [Line2D([0],
                                  [0],
                                  color=cmap(7),
                                  lw=EDGE_WIDTH,
                                  ls=':',
                                  label="Idle Movement")] + [Patch(facecolor=cmap(i),
                                                                   label=f"Tool {i}") for i in tools] + [Line2D([0],
                                                                                                                [0],
                                                                                                                color="black",
                                                                                                                lw=2 * EDGE_WIDTH,
                                                                                                                ls=style_options[i % len(
                                                                                                                    style_options)],
                                                                                                                label=f"Config {i + 1}") for i in configs]
        colors = nx.get_edge_attributes(G, 'color').values()
        widths = nx.get_edge_attributes(G, 'width').values()
        styles = [':' if i == -1 else style_options[i % len(style_options)]
                  for i in nx.get_edge_attributes(G, 'style').values()]

        nx.draw_networkx(
            G,
            pos,
            node_size=NODE_SIZE,
            font_size=FONT_SIZE,
            style=list(styles),
            edge_color=colors,
            edge_cmap=cmap,
            width=list(widths),
            connectionstyle=connectionstyle)

        plt.legend(handles=legend_elements)
        plt.savefig(path)
        plt.close()



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import final
from collections import deque
import json
from copy import deepcopy

from quark.Metrics import Metrics


class BenchmarkRecord:
    """
    The BenchmarkRecord class contains all the Metric instances and additional general information
    generated by a single benchmark run.
    """

    # pylint: disable=R0917
    def __init__(self, benchmark_backlog_item_number: int, timestamp: str, git_revision_number: str,
                 git_uncommitted_changes: str, repetition: int, total_repetitions: int):
        """
        Constructor method for BenchmarkRecord.

        :param benchmark_backlog_item_number: Number of the item in the benchmark backlog
        :param timestamp: Timestamp of the benchmark run
        :param git_revision_number: Git revision number during the benchmark run
        :param git_uncommitted_changes: Indication if there were uncommitted changes during the benchmark run
        :param repetition: Number of current repetitions of the benchmark run
        :param total_repetitions: Number of total repetitions of the benchmark run
        """
        self.benchmark_backlog_item_number = benchmark_backlog_item_number
        self.timestamp = timestamp
        self.git_revision_number = git_revision_number
        self.git_uncommitted_changes = git_uncommitted_changes
        self.repetition = repetition
        self.total_repetitions = total_repetitions
        self.total_time = None
        self.total_time_unit = "ms"
        self.linked_list_metrics = deque()

    @final
    def append_module_record_right(self, module_record: Metrics) -> None:
        """
        Adds Metrics instance to the end of the linked list.

        :param module_record: Metrics instance which should be appended to the end of the linked list
        """
        self.linked_list_metrics.append(module_record)

    @final
    def append_module_record_left(self, module_record: Metrics) -> None:
        """
        Adds Metrics instance to the beginning of the linked list.

        :param module_record: Metrics instance which should be appended to the beginning of the linked list
        """
        self.linked_list_metrics.appendleft(module_record)

    @final
    def sum_up_times(self) -> None:
        """
        Sums up the recording timings.
        """
        self.total_time = sum(item.total_time for item in self.linked_list_metrics)

    @final
    def hash_config(self, llist: deque) -> dict:
        """
        Recursively traverses through linked list and returns a dictionary with the next module's name, config,
        and subsequent submodule(s).

        :param llist: Linked list
        :return: Dictionary with the name and config of the first module in the linked list and subsequent submodule(s)
        """
        next_item: Metrics = llist.popleft()
        return {
            "module_name": next_item.module_name,
            "module_config": next_item.module_config,
            "submodule": self.hash_config(llist) if llist else {}
        }

    @final
    def start_hash_config(self) -> int:
        """
        Recursively creates dictionary containing the name of the module and its config for all modules in a chain.
        Then generates hash with it.

        :return: Hash of the benchmark run config
        """
        list_copy = deepcopy(self.linked_list_metrics)
        # Hash assumes that all keys are strings!
        return hash(json.dumps(self.hash_config(list_copy), sort_keys=True))

    @final
    def linked_list_to_dict(self, llist: deque, module_level: int = 0) -> dict:
        """
        Recursively traverses through linked list and adds the items of the Metrics objects to one single dictionary.

        :param llist: Linked list
        :param module_level: Current level in chain (starts at 0)
        :return: Dictionary with the module, its level, and its submodule(s)
        """
        next_item: Metrics = llist.popleft()
        return {
            **next_item.get(),
            "module_level": module_level,
            "submodule": self.linked_list_to_dict(llist, module_level + 1) if llist else {}
        }

    @final
    def start_linked_list_to_dict(self) -> dict:
        """
        Helper function to start linked_list_to_dict function which merges the various Metrics objects
        to one dictionary.

        :return: Resulting dictionary of linked_list_to_dict
        """
        list_copy = deepcopy(self.linked_list_metrics)
        return self.linked_list_to_dict(list_copy)

    @final
    def get(self) -> dict:
        """
        Returns a dictionary containing all benchmark information and a nested dictionary in which each level
        contains the metrics of the respective module.

        :return: Dictionary containing all the records of the benchmark
        """
        return {
            "benchmark_backlog_item_number": self.benchmark_backlog_item_number,
            "timestamp": self.timestamp,
            "config_hash": self.start_hash_config(),
            "total_time": self.total_time,
            "total_time_unit": self.total_time_unit,
            "git_revision_number": self.git_revision_number,
            "git_uncommitted_changes": self.git_uncommitted_changes,
            "repetition": self.repetition,
            "total_repetitions": self.total_repetitions,
            "module": self.start_linked_list_to_dict()
        }

    @final
    def copy(self) -> "BenchmarkRecord":
        """
        Returns a copy of itself.

        :return: Return copy of itself
        """
        return deepcopy(self)


class BenchmarkRecordStored:
    """
    This class can be used to store the BenchmarkRecord of a previous QUARK run as read from results.json.
    It is a simple wrapper with the purpose to provide the same interface to the BenchmarkManager as the
    BenchmarkRecord does.
    """

    def __init__(self, record: dict):
        """
        Constructor method for BenchmarkRecordStored.

        :param record: the record as dictionary
        """
        self.record = record

    def get(self) -> dict:
        """
        Simply returns the dictionary as given to the constructor.

        :return: Dictionary as given to the constructor
        """
        return self.record

    def sum_up_times(self) -> None:
        """
        Dummy implementation which does nothing.
        """
        pass



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import pickle
from typing import TypedDict

import networkx as nx
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

from quark.modules.applications.Application import Core
from quark.modules.applications.optimization.Optimization import Optimization
from quark.modules.applications.optimization.MIS.data.graph_layouts import generate_hexagonal_graph
from quark.utils import start_time_measurement, end_time_measurement

# Define R_rydberg
R_rydberg = 9.75


class MIS(Optimization):
    """
    The maximum independent set (MIS) problem is a combinatorial optimization problem that seeks to find the largest
    subset of vertices in a graph such that no two vertices are adjacent. MIS has numerous application in computer
    science, network design, resource allocation, and even in physics, where finding optimal configurations can
    solve fundamental problems related to stability and energy minimization.

    In a graph, the maximum independent set represents a set of nodes such that no two nodes share an edge. This
    property makes it a key element in various optimization scenarios. Due to the problem's combinatorial nature,
    it becomes computationally challenging, especially for large graphs, often requiring heuristic or approximate
    solutions.

    In the context of QUARK, we employ quantum-inspired approaches and state-of-the-art classical algorithms to
    tackle the problem. The graph is generated based on user-defined parameters such as size, spacing, and
    filling fraction, which affect the complexity and properties of the generated instance.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("MIS")
        self.submodule_options = ["QIRO", "NeutralAtom"]
        # TODO add more solvers like classical heuristics, VQE, QAOA, etc.
        self.depending_parameters = True
        self.graph = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return []

    def get_solution_quality_unit(self) -> str:
        """
        Returns the unit of measurement for solution quality.

        :return: The unit of measure for solution quality
        """
        return "Set size"

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "QIRO":
            from quark.modules.applications.optimization.MIS.mappings.QIRO import QIRO  # pylint: disable=C0415
            return QIRO()
        elif option == "NeutralAtom":
            from quark.modules.applications.optimization.MIS.mappings.NeutralAtom import NeutralAtom  # pylint: disable=C0415
            return NeutralAtom()
        else:
            raise NotImplementedError(f"Mapping Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application.

        :return: Configuration dictionary for this application
        .. code-block:: python

        return {
                "size": {
                    "values": [1, 5, 10, 15],
                    "custom_input": True,
                    "allow_ranges": True,
                    "postproc": int,
                    "description": "How large should your graph be?"
                },
                "graph_type": {
                    "values": ["hexagonal", "erdosRenyi"],
                    "postproc": str,
                    "description": "Do you want a hexagonal or an Erdos-Renyi graph?",
                    "depending_submodule": True
                }
            }
        """
        return {
            "size": {
                "values": [1, 5, 10, 15],
                "custom_input": True,
                "allow_ranges": True,
                "postproc": int,
                "description": "How large should your graph be?"
            },
            "graph_type": {
                "values": ["hexagonal", "erdosRenyi"],
                "postproc": str,
                "description": "Do you want a hexagonal or an Erdos-Renyi graph?",
                "depending_submodule": True
            }
        }

    def get_available_submodules(self, option: list) -> list:
        """
        Changes mapping options based on selection of graphs.

        :param option: List of chosen graph type
        :return: List of available submodules
        """
        if option == ["hexagonal"]:
            return ["QIRO", "NeutralAtom"]
        else:
            return ["QIRO"]

    def get_depending_parameters(self, option: str, config: dict) -> dict:
        """
        Returns parameters necessary for chosen problem option.

        :param option: The chosen option
        :param config: The current config
        :return: The parameters for the given option
        """

        more_params = {
            "filling_fraction": {
                "values": [x / 10 for x in range(2, 11, 2)],
                "custom_input": True,
                "allow_ranges": True,
                "postproc": float,
                "description": "What should be the filling fraction of the hexagonal graph / p of erdosRenyi graph?"
            }}
        if option == "QIRO":
            more_params["seed"] = {
                "values": ["No"],
                "custom_input": True,
                "description": "Do you want to set a seed? If yes, please set an integer number"
            }

        elif option == "NeutralAtom":
            pass  # No additional parameters needed at the moment
        else:
            raise NotImplementedError(f"Option {option} not implemented")
        if "hexagonal" in config["graph_type"]:
            more_params["spacing"] = {
                "values": [x / 10 for x in range(3, 11, 2)],
                "custom_input": True,
                "allow_ranges": True,
                "postproc": float,
                "description": "How much space do you want between your nodes, relative to Rydberg distance?"
            }
        param_to_return = {}
        for key, value in more_params.items():
            if key not in config:
                param_to_return[key] = value

        return param_to_return

    class Config(TypedDict):
        """
        Configuration attributes for generating an MIS problem.

        Attributes:
            size (int): The number of nodes in the graph.
            spacing (float): The spacing between nodes in the graph.
            filling_fraction (float): The fraction of available places in the lattice filled with nodes
        """
        size: int
        spacing: float
        filling_fraction: float

    def generate_problem(self, config: Config) -> nx.Graph:
        """
        Generates a graph to solve the MIS problem for.

        :param config: Config specifying the size and connectivity for the problem
        :return: Networkx graph representing the problem
        """
        if config is None:
            logging.warning("No config provided, using default values: graph_type='hexagonal', size=3, spacing=1,"
                            "filling_fraction=0.5")
            config = {"graph_type": "hexagonal", "size": 3, "spacing": 1, "filling_fraction": 0.5}

        graph_type = config.get('graph_type')
        size = config.get('size')
        filling_fraction = config.get('filling_fraction')

        if graph_type == "erdosRenyi":
            gseed = config.get("seed")

            if gseed == "No":
                graph = nx.erdos_renyi_graph(size, filling_fraction)

            else:
                try:
                    gseed = int(gseed)
                except ValueError:
                    logging.warning(f"Please select an integer number as seed for the Erdos-Renyi graph instead of "
                                    f"'{gseed}'. The seed is instead set to 0.")
                    gseed = 0
                graph = nx.erdos_renyi_graph(size, filling_fraction, seed=gseed)
            logging.info("Created MIS problem with the nx.erdos_renyi graph method, with the following attributes:")
            logging.info(f" - Graph size: {size}")
            logging.info(f" - p: {filling_fraction}")
            logging.info(f" - seed: {gseed}")

        else:
            if config.get('spacing') is None:
                spacing = 0.5
            else:
                spacing = config.get('spacing')
            graph = generate_hexagonal_graph(n_nodes=size,
                                             spacing=spacing * R_rydberg,
                                             filling_fraction=filling_fraction)
            logging.info("Created MIS problem with the generate hexagonal graph method, with the following attributes:")
            logging.info(f" - Graph size: {size}")
            logging.info(f" - Spacing: {spacing * R_rydberg}")
            logging.info(f" - Filling fraction: {filling_fraction}")

        self.graph = graph
        return graph.copy()

    def validate(self, solution: list) -> tuple[bool, float]:
        """
        Checks if the solution is an independent set.

        :param solution: List containing the nodes of the solution
        :return: Boolean whether the solution is valid and time it took to validate
        """
        start = start_time_measurement()
        is_valid = True

        nodes = list(self.graph.nodes())
        edges = list(self.graph.edges())

        # Check if the solution is independent
        is_independent = all((u, v) not in edges for u, v in edges if u in solution and v in solution)
        if is_independent:
            logging.info("The solution is independent.")
        else:
            logging.warning("The solution is not independent.")
            is_valid = False

        # Check if the solution is a set
        solution_set = set(solution)
        is_set = len(solution_set) == len(solution)
        if is_set:
            logging.info("The solution is a set.")
        else:
            logging.warning("The solution is not a set.")
            is_valid = False

        # Check if the solution is a subset of the original nodes
        is_subset = all(node in nodes for node in solution)
        if is_subset:
            logging.info("The solution is a subset of the problem.")
        else:
            logging.warning("The solution is not a subset of the problem.")
            is_valid = False

        return is_valid, end_time_measurement(start)

    def evaluate(self, solution: list) -> tuple[int, float]:
        """
        Calculates the size of the solution.

        :param solution: List containing the nodes of the solution
        :return: Set size, time it took to calculate the set size
        """
        start = start_time_measurement()
        set_size = len(solution)

        logging.info(f"Size of solution: {set_size}")

        return set_size, end_time_measurement(start)

    def save(self, path: str, iter_count: int) -> None:
        """
        Saves the generated problem graph to a file.

        :param path: Path to save the problem graph
        :param iter_count: Iteration count for file versioning
        """
        with open(f"{path}/graph_iter_{iter_count}.gpickle", "wb") as file:
            pickle.dump(self.graph, file, pickle.HIGHEST_PROTOCOL)

    def visualize_solution(self, processed_solution: list[int], path: str):
        """
        Plot the problem graph with the solution nodes highlighted

        :param processed_solution: The solution already processed by :func:`process_solution`, a list of visited node IDs in order of being visited.
        :param path: File path for the plot
        :returns: None
        """
        NODE_SIZE = 300   # Default=300
        EDGE_WIDTH = 1.0  # Default=1.0
        FONT_SIZE = 12    # Default=12
        COLOR_INCLUDED = "red"
        COLOR_EXCLUDED = "gray"

        G = self.graph
        included_nodes = [node for node in G.nodes() if node in processed_solution]
        excluded_nodes = [node for node in G.nodes() if node not in processed_solution]
        pos = nx.circular_layout(G)
        included_pos = {n: n for n, _ in pos.items() if n in processed_solution}
        excluded_pos = {n: n for n, _ in pos.items() if n not in processed_solution}
        legend_elements = [
            Line2D(
                [0],
                [0],
                marker='o',
                ls="None",
                label="Included",
                markerfacecolor=COLOR_INCLUDED,
                markeredgewidth=0,
                markersize=10),
            Line2D(
                [0],
                [0],
                marker='o',
                ls="None",
                label="Excluded",
                markerfacecolor=COLOR_EXCLUDED,
                markeredgewidth=0,
                markersize=10)
        ]

        nx.draw_networkx_nodes(G, pos, nodelist=included_nodes, node_size=NODE_SIZE, node_color=COLOR_INCLUDED)
        nx.draw_networkx_nodes(G, pos, nodelist=excluded_nodes, node_size=NODE_SIZE, node_color=COLOR_EXCLUDED)
        nx.draw_networkx_labels(G, pos, included_pos, font_size=FONT_SIZE, font_weight="bold")
        nx.draw_networkx_labels(G, pos, excluded_pos, font_size=FONT_SIZE)
        nx.draw_networkx_edges(G, pos, width=EDGE_WIDTH)

        plt.legend(handles=legend_elements)
        plt.savefig(path)
        plt.close()



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import itertools
from collections import defaultdict
from typing import TypedDict
import logging

import networkx as nx

from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement


class QUBO(Mapping):
    """
    QUBO formulation for the PVC.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Annealer"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dictionaries with requirements of this module
        """
        return [{"name": "networkx", "version": "3.4.2"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dictionary containing parameter options
        .. code-block:: python

            return {
                    "lagrange_factor": {
                        "values": [0.75, 1.0, 1.25],
                        "description": "By which factor would you like to multiply your Lagrange?"
                    }
                }

        """
        return {
            "lagrange_factor": {
                "values": [0.75, 1.0, 1.25],
                "description": "By which factor would you like to multiply your Lagrange?"
            }
        }

    class Config(TypedDict):
        """
        Configuration attributes of QUBO mapping.

        Attributes:
             lagrange_factor (float): Factor to multiply the Langrange.

        """
        lagrange_factor: float

    def map(self, problem: nx.Graph, config: Config) -> tuple[dict, float]:
        """
        Maps the networkx graph to a QUBO formulation.

        :param problem: Networkx graph representing the PVC problem
        :param config: Config dictionary with the mapping configuration
        :return: Tuple containing the QUBO dictionary and the time it took to map it
        """
        # Inspired by https://dnx.readthedocs.io/en/latest/_modules/dwave_networkx/algorithms/tsp.html
        start = start_time_measurement()
        lagrange_factor = config['lagrange_factor']

        # Estimate lagrange if not provided
        n = problem.number_of_nodes()
        timesteps = int((n - 1) / 2 + 1)

        # Get the number of different configs and tools
        config = [x[2]['c_start'] for x in problem.edges(data=True)]
        config = list(set(config + [x[2]['c_end'] for x in problem.edges(data=True)]))

        tool = [x[2]['t_start'] for x in problem.edges(data=True)]
        tool = list(set(tool + [x[2]['t_end'] for x in problem.edges(data=True)]))

        if problem.number_of_edges() > 0:
            weights = [x[2]['weight'] for x in problem.edges(data=True)]
            weights = list(filter(lambda a: a != max(weights), weights))
            lagrange = sum(weights) / len(weights) * timesteps
        else:
            lagrange = 2

        lagrange *= lagrange_factor
        logging.info(f"Selected lagrange is: {lagrange}")

        if n in (1, 2) or len(problem.edges) < n * (n - 1) // 2:
            msg = "graph must be a complete graph with at least 3 nodes or empty"
            raise ValueError(msg)

        # Creating the QUBO
        q = defaultdict(float)

        # We need to implement the following constrains:
        # Only visit 1 node of each seam
        # Don`t visit nodes twice (even if their config/tool is different)
        # We only need to visit base node at the once since this path from last node to base node is unique anyway

        # Constraint to only visit a node/seam once
        for node in problem:  # for all nodes in the graph
            for pos_1 in range(timesteps):  # for number of timesteps
                for t_start in tool:
                    for c_start in config:
                        q[((node, c_start, t_start, pos_1), (node, c_start, t_start, pos_1))] -= lagrange
                        for t_end in tool:
                            # For all configs and tools
                            for c_end in config:
                                if c_start != c_end or t_start != t_end:
                                    q[((node, c_start, t_start, pos_1), (node, c_end, t_end, pos_1))] += 1.0 * lagrange
                                for pos_2 in range(pos_1 + 1, timesteps):
                                    # Penalize visiting same node again in another timestep
                                    q[((node, c_start, t_start, pos_1), (node, c_end, t_end, pos_2))] += 2.0 * lagrange
                                    # Penalize visiting other node of same seam
                                    if node != (0, 0):
                                        # (0,0) is the base node, it is not a seam
                                        # Get the other nodes of the same seam
                                        other_seam_nodes = [
                                            x for x in problem.nodes if x[0] == node[0] and x[1] != node
                                        ]
                                        for other_seam_node in other_seam_nodes:
                                            # Penalize visiting other node of same seam
                                            q[((node, c_start, t_start, pos_1),
                                               (other_seam_node, c_end, t_end, pos_2))] += 2.0 * lagrange

        # Constraint to only visit a single node in a single timestep
        for pos in range(timesteps):
            for node_1 in problem:
                for t_start in tool:
                    for c_start in config:
                        q[((node_1, c_start, t_start, pos), (node_1, c_start, t_start, pos))] -= lagrange
                        for t_end in tool:
                            for c_end in config:
                                for node_2 in set(problem) - {node_1}:  # for all nodes except node1 -> node1
                                    q[((node_1, c_start, t_start, pos), (node_2, c_end, t_end, pos))] += lagrange

        # Objective that minimizes distance
        for u, v in itertools.combinations(problem.nodes, 2):
            for pos in range(timesteps):
                for t_start in tool:
                    for t_end in tool:
                        for c_start in config:
                            for c_end in config:
                                nextpos = (pos + 1) % timesteps
                                edge_u_v = next(
                                    item for item in list(problem[u][v].values())
                                    if item["c_start"] == c_start and item["t_start"] == t_start and
                                    item["c_end"] == c_end and item["t_end"] == t_end
                                )
                                # Since it is the other direction we switch start and end of tool and config
                                edge_v_u = next(
                                    item for item in list(problem[v][u].values())
                                    if item["c_start"] == c_end and item["t_start"] == t_end and
                                    item["c_end"] == c_start and item["t_end"] == t_start
                                )
                                # Going from u -> v
                                q[((u, c_start, t_start, pos), (v, c_end, t_end, nextpos))] += edge_u_v['weight']
                                # Going from v -> u
                                q[((v, c_end, t_end, pos), (u, c_start, t_start, nextpos))] += edge_v_u['weight']

        logging.info("Created Qubo")

        return {"Q": q}, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "Annealer":
            from quark.modules.solvers.Annealer import Annealer  # pylint: disable=C0415
            return Annealer()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import networkx as nx
import pickle
import tsplib95

# Source http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/tsp/
filename = "dsj1000.tsp"


def main():
    """
    Load a TSP problem, remove unnecessary edges, and save the reference graph.
    """
    print(f"Loading {filename}")

    # Load the problem from .tsp file
    problem = tsplib95.load(filename)
    graph = problem.get_graph()

    # We don't needed edges from e.g. node0 -> node0
    for edge in graph.edges:
        if edge[0] == edge[1]:
            graph.remove_edge(edge[0], edge[1])

    print("Loaded graph:")
    print(nx.info(graph))

    with open("reference_graph.gpickle", "wb") as file:
        pickle.dump(graph, file, pickle.HIGHEST_PROTOCOL)

    print("Saved graph as reference_graph.gpickle")


if __name__ == '__main__':
    main()



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union, TypedDict
from itertools import combinations
from scipy.special import binom

from quark.modules.applications.qml.generative_modeling.circuits.CircuitGenerative import CircuitGenerative
from quark.modules.applications.qml.generative_modeling.mappings.LibraryQiskit import LibraryQiskit
from quark.modules.applications.qml.generative_modeling.mappings.LibraryPennylane import LibraryPennylane
from quark.modules.applications.qml.generative_modeling.mappings.PresetQiskitNoisyBackend import PresetQiskitNoisyBackend
from quark.modules.applications.qml.generative_modeling.mappings.CustomQiskitNoisyBackend import CustomQiskitNoisyBackend


class CircuitCopula(CircuitGenerative):
    """
    This class generates a library-agnostic gate sequence, i.e. a list containing information
    about the gates and the wires they act on. The marginal distributions generated by the copula
    are uniformaly distributed.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("DiscreteCopula")
        self.submodule_options = [
            "LibraryQiskit",
            "LibraryPennylane",
            "CustomQiskitNoisyBackend",
            "PresetQiskitNoisyBackend"
        ]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "scipy", "version": "1.12.0"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this Copula Circuit.

        :return: Dictionary of parameter options
        .. code-block:: python

            return {
                "depth": {
                    "values": [1, 2, 3, 4, 5],
                    "description": "What depth do you want?"
                },
            }

        """
        return {
            "depth": {
                "values": [1, 2, 3, 4, 5],
                "description": "What depth do you want?"
            },
        }

    def get_default_submodule(self, option: str) \
            -> Union[LibraryQiskit, LibraryPennylane, PresetQiskitNoisyBackend, CustomQiskitNoisyBackend]:
        """
        Returns the default submodule based on the given option.

        :param option: The submodule option to select
        :return: Instance of the selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "LibraryQiskit":
            return LibraryQiskit()
        if option == "LibraryPennylane":
            return LibraryPennylane()
        elif option == "PresetQiskitNoisyBackend":
            return PresetQiskitNoisyBackend()
        elif option == "CustomQiskitNoisyBackend":
            return CustomQiskitNoisyBackend()
        else:
            raise NotImplementedError(f"Option {option} not implemented")

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             depth: int

        """
        depth: int

    def generate_gate_sequence(self, input_data: dict, config: Config) -> dict:
        """
        Returns gate sequence of copula architecture.

        :param input_data: Collection of information from the previous modules
        :param config: Config specifying the number of qubits of the circuit
        :return: Dictionary including the gate sequence of the Copula Circuit
        """
        n_registers = input_data["n_registers"]
        n_qubits = input_data["n_qubits"]
        depth = config["depth"]
        n = n_qubits // n_registers

        gate_sequence = []
        for k in range(n):
            gate_sequence.append(["Hadamard", [k]])

        for j in range(n_registers - 1):
            for k in range(n):
                gate_sequence.append(["CNOT", [k, k + n * (j + 1)]])

        gate_sequence.append(["Barrier", None])

        shift = 0
        for _ in range(depth):
            for k in range(n):
                for j in range(n_registers):
                    gate_sequence.append(["RZ", [j * n + k]])
                    gate_sequence.append(["RX", [j * n + k]])
                    gate_sequence.append(["RZ", [j * n + k]])

            k = 3 * n + shift
            for i, j in combinations(range(n), 2):
                for layer in range(n_registers):
                    gate_sequence.append(["RXX", [layer * n + i, layer * n + j]])

                k += 1
            shift += 3 * n + int(binom(n, 2))

        gate_sequence.append(["Barrier", None])

        for k in range(n_qubits):
            gate_sequence.append(["Measure", [k, k]])

        output_dict = {
            "gate_sequence": gate_sequence,
            "circuit_name": "Copula",
            "n_qubits": n_qubits,
            "n_registers": n_registers,
            "depth": depth,
            "histogram_train": input_data["histogram_train"],
            "store_dir_iter": input_data["store_dir_iter"],
            "train_size": input_data["train_size"],
            "dataset_name": input_data["dataset_name"],
            "binary_train": input_data["binary_train"]
        }

        return output_dict



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import logging

import networkx as nx
import numpy as np
from dimod import qubo_to_ising

from quark.modules.applications.Mapping import Mapping, Core
from quark.modules.applications.optimization.PVC.mappings.QUBO import QUBO
from quark.utils import start_time_measurement, end_time_measurement


class Ising(Mapping):
    """
    Ising formulation for the PVC.

    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["QAOA", "PennylaneQAOA"]
        self.key_mapping = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dictionaries with requirements of this module
        """
        return [
            {"name": "networkx", "version": "3.4.2"},
            {"name": "numpy", "version": "1.26.4"},
            {"name": "dimod", "version": "0.12.18"},
            *QUBO.get_requirements()
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dictionary containing parameter options.
        .. code-block:: python

            return {
                    "lagrange_factor": {
                        "values": [0.75, 1.0, 1.25],
                        "description": "By which factor would you like to multiply your Lagrange?"
                    }
                }
        """
        return {
            "lagrange_factor": {
                "values": [0.75, 1.0, 1.25],
                "description": "By which factor would you like to multiply your Lagrange?"
            }
        }

    class Config(TypedDict):
        """
        Configuration attributes for Ising mapping.

        Attributes:
             lagrange_factor (float): Factor to multiply the Langrange.
        """
        lagrange_factor: float

    def map(self, problem: nx.Graph, config: Config) -> tuple[dict, float]:
        """
        Uses the PVC QUBO formulation and converts it to an Ising representation.

        :param problem: Networkx graph representing the PVC problem
        :param config: Config dictionary with the mapping configuration
        :return: Tuple containing a dictionary with the ising problem and time it took to map it
        """
        start = start_time_measurement()

        # Convert the PVC problem to QUBO
        qubo_mapping = QUBO()
        q, _ = qubo_mapping.map(problem, config)

        # Convert QUBO to ising using dimod
        t, j, _ = qubo_to_ising(q["Q"])

        # Extract unique configuration and tool attributes from the graph
        config = [x[2]['c_start'] for x in problem.edges(data=True)]
        config = list(set(config + [x[2]['c_end'] for x in problem.edges(data=True)]))

        tool = [x[2]['t_start'] for x in problem.edges(data=True)]
        tool = list(set(tool + [x[2]['t_end'] for x in problem.edges(data=True)]))

        # Initialize J matrix and mapping
        timesteps = int((problem.number_of_nodes() - 1) / 2 + 1)
        matrix_size = problem.number_of_nodes() * len(config) * len(tool) * timesteps
        j_matrix = np.zeros((matrix_size, matrix_size), dtype=float)
        self.key_mapping = {}

        # Map J values to a matrix representation
        index_counter = 0
        for key, value in j.items():
            if key[0] not in self.key_mapping:
                self.key_mapping[key[0]] = index_counter
                index_counter += 1
            if key[1] not in self.key_mapping:
                self.key_mapping[key[1]] = index_counter
                index_counter += 1
            u = self.key_mapping[key[0]]
            v = self.key_mapping[key[1]]
            j_matrix[u][v] = value

        return {"J": j_matrix, "t": np.array(list(t.values()))}, end_time_measurement(start)

    def reverse_map(self, solution: dict) -> tuple[dict, float]:
        """
        Maps the solution back to the representation needed by the PVC class for validation/evaluation.

        :param solution: Dictionary containing the solution
        :return: Tuple with the remapped solution and time it took to reverse map
        """
        start = start_time_measurement()
        logging.info(f"Key Mapping: {self.key_mapping}")

        result = {key: 1 if solution[self.key_mapping[key]] == 1 else 0 for key in self.key_mapping}

        return result, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "QAOA":
            from quark.modules.solvers.QAOA import QAOA  # pylint: disable=C0415
            return QAOA()
        if option == "PennylaneQAOA":
            from quark.modules.solvers.PennylaneQAOA import PennylaneQAOA  # pylint: disable=C0415
            return PennylaneQAOA()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import importlib
import logging
import os
import subprocess
import sys
import time
from typing import Union, Callable

import inquirer


def _get_instance_with_sub_options(options: list[dict], name: str) -> any:
    """
    Creates an instance of the QUARK module identified by class_name.

    :param options: Section of the QUARK module configuration including the submodules' information
    :param name: Name of the QUARK component to be initialized
    :return: New instance of the QUARK module
    """
    for opt in options:
        if name != opt["name"]:
            continue
        class_name = opt.get("class", name)
        clazz = _import_class(opt["module"], class_name, opt.get("dir"))
        sub_options = opt.get("submodules", None)

        # In case the class requires some arguments in its constructor
        # they can be defined in the "args" dict
        if "args" in opt and opt["args"]:
            instance = clazz(**opt["args"])
        else:
            instance = clazz()

        # _get_instance_with_sub_options is mostly called when using the --modules option,
        # so it makes sense to also
        # save the git revision of the given module, since it can be in a different git

        # Directory of this file
        utils_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)))
        module_dir = os.path.dirname(sys.modules[opt["module"]].__file__)

        # Only log git revision number, if the module is not in the same directory as the utils file
        if not os.path.commonprefix([utils_dir, module_dir]) == utils_dir:
            git_revision_number, git_uncommitted_changes = get_git_revision(module_dir)

            instance.metrics.add_metric_batch({
                "module_git_revision_number": git_revision_number,
                "module_git_uncommitted_changes": git_uncommitted_changes
            })

        # sub_options inherits 'dir'
        if sub_options and "dir" in opt:
            for sub_opt in sub_options:
                if "dir" not in sub_opt:
                    sub_opt["dir"] = opt["dir"]

        instance.sub_options = sub_options
        return instance

    logging.error(f"{name} not found in {options}")
    raise ValueError(f"{name} not found in {options}")


def _import_class(module_path: str, class_name: str, base_dir: str = None) -> any:
    """
    Helper function which allows to replace hard-coded imports of the form
    'import MyClass from path.to.mypkg' by calling _import_class('path.to.mypkg', 'MyClass').
    If base_dir is specified, its value will be added to the python search path,
    unless it's already contained in it.

    :param module_path: Python module path of the module containing the class to be imported
    :param class_name: Name of the class to be imported
    :return: Imported class object
    """

    # Make sure that base_dir is in the search path.
    # Otherwise, the module imported here might not find its libraries.
    if base_dir is not None and base_dir not in sys.path:
        logging.info(f"Append to sys.path: {base_dir}")
        sys.path.append(base_dir)
    logging.info(f"Import module {module_path}")
    module = importlib.import_module(module_path)
    return vars(module)[class_name]


def checkbox(key: str, message: str, choices: list) -> dict:
    """
    Wrapper method to avoid empty responses in checkboxes.

    :param key: Key for response dict
    :param message: Message for the user
    :param choices: Choices for the user
    :return: Dict with the response from the user
    """
    if len(choices) > 1:
        answer = inquirer.prompt([inquirer.Checkbox(key, message=message, choices=choices)])
    else:
        if len(choices) == 1:
            logging.info(f"Skipping asking for submodule"
                         f"since only 1 option ({choices[0]}) is available.")
        return {key: choices}

    if not answer[key]:
        logging.warning("You must check at least one box! Please try again!")
        return checkbox(key, message, choices)

    return answer


def get_git_revision(git_dir: str) -> tuple[str, str]:
    """
    Collects git revision number and checks if there are uncommitted changes
    to allow user to analyze which codebase was used.

    :param git_dir: Directory of the git repository
    :return: Tuple with git_revision_number, git_uncommitted_changes
    """
    try:
        # '-C', git_dir ensures that the following commands also work
        # when QUARK is started from other working directories
        git_revision_number = subprocess.check_output(
            ['git', '-C', git_dir, 'rev-parse', 'HEAD']).decode('ascii').strip()
        git_uncommitted_changes = bool(subprocess.check_output(
            ['git', '-C', git_dir, 'status', '--porcelain', '--untracked-files=no']).decode(
            'ascii').strip())

        logging.info(
            f"Codebase is based on revision {git_revision_number} and has "
            f"{'some' if git_uncommitted_changes else 'no'} uncommitted changes"
        )
    except Exception as e:
        logging.warning(f"Logging of git revision number not possible because of: {e}")
        git_revision_number = "unknown"
        git_uncommitted_changes = "unknown"

    return git_revision_number, git_uncommitted_changes


#autopep8: off
def _expand_paths(j: Union[dict, list], base_dir: str) -> Union[dict, list]:
    """
    Expands the paths given as value of the 'dir' attribute appearing in the QUARK modules
    configuration by joining base_dir with that path.

    :param j: The JSON to be adapted - expected to be a QUARK modules configuration or a part of it
    :param base_dir: The base directory to be used for path expansion
    :return: The adapted JSON
    """
    assert type(j) in [dict, list], f"unexpected type:{type(j)}"
    if type(j) is list:
        for entry in j:
            _expand_paths(entry, base_dir)
    else:
        for attr in j:
            if type(j[attr]) == "submodules":
                _expand_paths(j[attr], base_dir)
            elif attr == "dir":
                p = j[attr]
                if not os.path.isabs(p):
                    j[attr] = os.path.join(base_dir, p)
    return j


def start_time_measurement() -> float:
    """
    Starts a time measurement.

    :return: Starting point
    """
    return time.perf_counter()


def end_time_measurement(start: float) -> float:
    """
    Returns the result of the time measurement in milliseconds.

    :param start: Starting point for the measurement
    :return: Time elapsed in ms
    """
    end = time.perf_counter()
    return round((end - start) * 1000, 3)


def stop_watch(position: int = None) -> Callable:
    """
    Usage as decorator to measure time, e.g.:
    ```
    @stop_watch()
    def run(input_data,...):
        return processed_data
    ```
    results in valid:    
    ```
    processed_data, time_to_process = run(input,...)
    ```
    If the return value of the decorated function is of type tuple,
    the optional parameter `position` can be used to specify the position at which the
    measured time is to be inserted in the return tuple.

    :param position: The position at which the measured time gets inserted in the return tuple.
    If not specified the measured time will be appended to the original return value.
    :return: The wrapper function
    """
    def wrap(func):
        def wrapper(*args, **kwargs):
            start = start_time_measurement()
            return_value = func(*args, **kwargs)
            duration = end_time_measurement(start)
            if position is not None and isinstance(return_value, tuple):
                rv = list(return_value)
                rv.insert(position, duration)
                return tuple(rv)
            return return_value, duration
        return wrapper
    return wrap



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from typing import TypedDict
import logging

import networkx as nx
import numpy as np
from dimod import qubo_to_ising
from more_itertools import locate
from qiskit_optimization.applications import Tsp
from qiskit_optimization.converters import QuadraticProgramToQubo

from quark.modules.applications.Mapping import Mapping, Core
from quark.modules.applications.optimization.TSP.mappings.QUBO import QUBO
from quark.utils import start_time_measurement, end_time_measurement


class Ising(Mapping):
    """
    Ising formulation for the TSP.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["QAOA", "PennylaneQAOA", "QiskitQAOA"]
        self.key_mapping = None
        self.graph = None
        self.config = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "networkx", "version": "3.4.2"},
            {"name": "numpy", "version": "1.26.4"},
            {"name": "dimod", "version": "0.12.18"},
            {"name": "more-itertools", "version": "10.5.0"},
            {"name": "qiskit-optimization", "version": "0.6.1"},
            *QUBO.get_requirements()
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dictionary containing parameter options.
        .. code-block:: python

            return {
                    "lagrange_factor": {
                        "values": [0.75, 1.0, 1.25],
                        "description": "By which factor would you like to multiply your lagrange?"
                    },
                    "mapping": {
                        "values": ["ocean", "qiskit"],
                        "description": "Which Ising formulation of the TSP problem should be used?"
                    }
                }
        """
        return {
            "lagrange_factor": {
                "values": [0.75, 1.0, 1.25],
                "description": "By which factor would you like to multiply your lagrange?"
            },
            "mapping": {
                "values": ["ocean", "qiskit"],
                "description": "Which Ising formulation of the TSP problem should be used?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             lagrange_factor: float
             mapping: str

        """
        lagrange_factor: float
        mapping: str

    def map(self, problem: nx.Graph, config: Config) -> tuple[dict, float]:
        """
        Maps the networkx graph to an Ising formulation.

        :param problem: Networkx graph
        :param config: Config with the parameters specified in Config class
        :return: Dict with Ising, time it took to map it
        """
        self.graph = problem
        self.config = config
        # Call mapping function defined in configuration
        mapping = self.config["mapping"]
        if mapping == "ocean":
            return self._map_ocean(problem, config)
        elif mapping == "qiskit":
            return self._map_qiskit(problem, config)
        else:
            logging.error(f"Unknown mapping {mapping}.")
            raise ValueError(f"Unknown mapping {mapping}.")

    def _map_ocean(self, graph: nx.Graph, config: Config) -> tuple[dict, float]:
        """
        Use D-Wave/Ocean TSP QUBO/Ising model.

        :param graph: Networkx graph
        :param config: Config with the parameters specified in Config class
        :return: Dict with the Ising, time it took to map it
        """
        start = start_time_measurement()
        qubo_mapping = QUBO()
        q, _ = qubo_mapping.map(graph, config)
        t, j, _ = qubo_to_ising(q["Q"])

        # Convert ISING dict to matrix
        timesteps = graph.number_of_nodes()
        matrix_size = graph.number_of_nodes() * timesteps
        j_matrix = np.zeros((matrix_size, matrix_size), dtype=float)

        self.key_mapping = {}
        index_counter = 0

        for key, value in j.items():
            if key[0] not in self.key_mapping:
                self.key_mapping[key[0]] = index_counter
                index_counter += 1
            if key[1] not in self.key_mapping:
                self.key_mapping[key[1]] = index_counter
                index_counter += 1
            u = self.key_mapping[key[0]]
            v = self.key_mapping[key[1]]
            j_matrix[u][v] = value

        return {"J": j_matrix, "t": np.array(list(t.values())), "J_dict": j}, end_time_measurement(start)

    @staticmethod
    def _map_qiskit(graph: nx.Graph, config: Config) -> tuple[dict, float]:
        """
        Use Ising Mapping of Qiskit Optimize:
        TSP class: https://qiskit.org/documentation/optimization/stubs/qiskit_optimization.applications.Tsp.html
        Example notebook: https://qiskit.org/documentation/tutorials/optimization/6_examples_max_cut_and_tsp.html

        :param graph: Networkx graph
        :param config: Config with the parameters specified in Config class
        :return: Dict with the Ising, time it took to map it
        """
        start = start_time_measurement()
        tsp = Tsp(graph)
        qp = tsp.to_quadratic_program()
        logging.info(qp.export_as_lp_string())
        qp2qubo = QuadraticProgramToQubo()
        qubo = qp2qubo.convert(qp)
        qubitOp, _ = qubo.to_ising()

        # Reverse generate J and t out of qubit PauliSumOperator from qiskit
        t_matrix = np.zeros(qubitOp.num_qubits, dtype=complex)
        j_matrix = np.zeros((qubitOp.num_qubits, qubitOp.num_qubits), dtype=complex)
        pauli_list = qubitOp.to_list()

        for pauli_str, coeff in pauli_list:
            pauli_str_list = list(pauli_str)
            index_pos_list = list(locate(pauli_str_list, lambda a: a == 'Z'))
            if len(index_pos_list) == 1:
                t_matrix[index_pos_list[0]] = coeff
            elif len(index_pos_list) == 2:
                j_matrix[index_pos_list[0]][index_pos_list[1]] = coeff

        return {"J": j_matrix, "t": t_matrix}, end_time_measurement(start)

    def reverse_map(self, solution: any) -> tuple[dict, float]:
        """
        Maps the solution back to the representation needed by the TSP class for validation/evaluation.

        :param solution: List or array containing the solution
        :return: Solution mapped accordingly, time it took to map it
        """
        start = start_time_measurement()
        if -1 in solution:  # ising model output from Braket QAOA
            solution = self._convert_ising_to_qubo(solution)
        elif self.config["mapping"] == "ocean":
            logging.debug("Flip bits in solutions to unify different mappings")
            solution = self._flip_bits_in_bitstring(solution)

        logging.info(f"Best Bitstring: {solution}")
        n = self.graph.number_of_nodes()

        result = {}
        if self.key_mapping is None:
            # Node indexes in graph are used as index in qubits
            it = np.nditer(solution, flags=['multi_index'])
            for x in it:
                logging.debug(f"{x}, {it.multi_index}")
                idx = it.multi_index[0]
                result[(int(idx / n), int(idx % n))] = x
        else:
            logging.debug("Using key Mapping: {self.key_mapping}")
            for key, value in self.key_mapping.items():
                result[key] = 1 if solution[value] == 1 else 0

        return result, end_time_measurement(start)

    @staticmethod
    def _flip_bits_in_bitstring(solution: any) -> any:
        """
        Flip bits in the solution bitstring to unify different mappings.

        :param solution: Solution bitstring
        :return: Flipped solution bitstring
        """
        solution = np.array(solution)
        with np.nditer(solution, op_flags=['readwrite']) as it:
            for x in it:
                x[...] = 1 - x

        return solution

    @staticmethod
    def _convert_ising_to_qubo(solution: any) -> any:
        """
        Convert Ising model output to QUBO format.

        :param solution: Ising model output
        :return: QUBO format solution
        """
        solution = np.array(solution)
        with np.nditer(solution, op_flags=['readwrite']) as it:
            for x in it:
                if x == -1:
                    x[...] = 0

        return solution

    def get_default_submodule(self, option: str) -> Core:
        """
        Get the default submodule based on the given option.

        :param option: Submodule option
        :return: Corresponding submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "QAOA":
            from quark.modules.solvers.QAOA import QAOA  # pylint: disable=C0415
            return QAOA()
        elif option == "PennylaneQAOA":
            from quark.modules.solvers.PennylaneQAOA import PennylaneQAOA  # pylint: disable=C0415
            return PennylaneQAOA()
        elif option == "QiskitQAOA":
            from quark.modules.solvers.QiskitQAOA import QiskitQAOA  # pylint: disable=C0415
            return QiskitQAOA()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import pandas as pd

from quark.modules.applications.qml.generative_modeling.transformations.Transformation import Transformation
from quark.modules.applications.qml.generative_modeling.circuits.CircuitCopula import CircuitCopula


class PIT(Transformation):  # pylint disable=R0902
    """
    The transformation of the original probability distribution to
    the distribution of its uniformly distributed cumulative marginals is known as the copula.
    """

    def __init__(self):
        super().__init__("PIT")
        self.submodule_options = ["CircuitCopula"]
        self.reverse_epit_lookup = None
        self.transform_config = None
        self.n_qubits = None
        self.dataset = None
        self.dataset_name = None
        self.grid_shape = None
        self.histogram_train = None
        self.histogram_train_original = None
        self.histogram_transformed = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "numpy", "version": "1.26.4"},
            {"name": "pandas", "version": "2.2.3"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns empty dict as this transformation has no configurable settings.

        :return: Empty dict
        """
        return {}

    def get_default_submodule(self, option: str) -> CircuitCopula:
        if option == "CircuitCopula":
            return CircuitCopula()
        else:
            raise NotImplementedError(f"Circuit Option {option} not implemented")

    def transform(self, input_data: dict, config: dict) -> dict:
        """
        Transforms the input dataset using PIT transformation and computes histograms
        of the training dataset in the transformed space.

        :param input_data: Dataset
        :param config: Config with the parameters specified in Config class
        :return: Dict with PIT transformation, time it took to map it
        """
        # TODO: PIT.transform is almost identical to MinMax.transform -> function should be moved to Transformation.py
        self.dataset_name = input_data["dataset_name"]
        self.dataset = input_data["dataset"]
        self.n_qubits = input_data["n_qubits"]
        self.grid_shape = int(2 ** (self.n_qubits // 2))
        n_registers = self.dataset.shape[-1]

        # Calculate ranges for the original dataset and the transformed dataset
        ranges_original = np.column_stack((np.min(self.dataset, axis=0), np.max(self.dataset, axis=0)))
        transformed_dataset = self.fit_transform(self.dataset)
        ranges_transformed = np.column_stack((np.min(transformed_dataset, axis=0), np.max(transformed_dataset, axis=0)))

        # Compute histogram for the transformed dataset
        transformed_histogram_grid = np.histogramdd(
            transformed_dataset,
            bins=self.grid_shape,
            range=ranges_transformed)[0]
        histogram_transformed_1d = transformed_histogram_grid.flatten()
        self.histogram_transformed = histogram_transformed_1d / np.sum(histogram_transformed_1d)

        solution_space = np.zeros(len(transformed_dataset), dtype=int)

        # Initialize a variable to keep track of the current position in the result_array
        position = 0
        value = 0
        for count in histogram_transformed_1d:
            if count > 0:
                solution_space[position:position + int(count)] = value
                position += int(count)
            value += 1

        binary_strings = [np.binary_repr(x, width=self.n_qubits) for x in solution_space]
        # Convert the binary strings to a NumPy array of integers
        binary_transformed = np.array([list(map(int, s)) for s in binary_strings])

        # Compute histogram for the original dataset
        learned_histogram = np.histogramdd(self.dataset, bins=self.grid_shape, range=ranges_original)
        self.histogram_train_original = learned_histogram[0] / np.sum(learned_histogram[0])

        # Compute histogram for the transformed dataset
        train_histogram = np.histogramdd(transformed_dataset, bins=self.grid_shape, range=ranges_transformed)
        histogram_train = train_histogram[0] / np.sum(train_histogram[0])
        self.histogram_train = histogram_train.flatten()

        self.transform_config = {
            "histogram_train": self.histogram_train,
            "binary_train": binary_transformed,
            "dataset_name": self.dataset_name,
            "n_registers": n_registers,
            "n_qubits": self.n_qubits,
            "store_dir_iter": input_data["store_dir_iter"],
            "train_size": input_data["train_size"],
            "transformed_dataset": transformed_dataset
        }

        return self.transform_config

    def reverse_transform(self, input_data: dict) -> dict:
        """
        Transforms the solution back to the representation needed for validation/evaluation.

        :param input_data: Dictionary containing the solution
        :return: Dictionary with solution transformed accordingly
        """
        depth = input_data["depth"]
        architecture_name = input_data["architecture_name"]
        n_qubits = input_data["n_qubits"]
        n_registers = self.transform_config["n_registers"]
        kl_best_transformed = min(input_data["KL"])
        best_results = input_data["best_sample"]
        circuit_transpiled = input_data['circuit_transpiled']

        array_bins = self.compute_discretization_efficient(n_qubits, n_registers)
        transformed_samples = self.generate_samples_efficient(best_results, array_bins, n_registers, noisy=True)

        # Calculate ranges for the transformed samples
        ranges_transformed = np.column_stack((np.min(transformed_samples, axis=0), np.max(transformed_samples, axis=0)))

        # Compute histogram for the transformed samples
        learned_histogram = np.histogramdd(transformed_samples, bins=self.grid_shape, range=ranges_transformed)
        histogram_generated_transformed = learned_histogram[0] / np.sum(learned_histogram[0])
        histogram_generated_transformed = histogram_generated_transformed.flatten()

        original_samples = self.inverse_transform(transformed_samples)

        # Calculate ranges for the original samples
        ranges_original = np.column_stack((np.min(original_samples, axis=0), np.max(original_samples, axis=0)))

        # Compute histogram for the original samples
        learned_histogram = np.histogramdd(original_samples, bins=self.grid_shape, range=ranges_original)
        histogram_generated_original = learned_histogram[0] / np.sum(learned_histogram[0])
        histogram_generated_original = histogram_generated_original.flatten()

        best_parameter = input_data["best_parameter"]

        reverse_config_trans = {
            "generated_samples": best_results,
            "transformed_samples": transformed_samples,
            "depth": depth,
            "architecture_name": architecture_name,
            "dataset_name": self.dataset_name,
            "n_qubits": n_qubits,
            "best_parameter": best_parameter,
            "histogram_train_original": self.histogram_train_original,
            "histogram_train": self.histogram_train,
            "histogram_generated_original": histogram_generated_original,
            "histogram_generated": histogram_generated_transformed,
            "KL_best_transformed": kl_best_transformed,
            "store_dir_iter": input_data["store_dir_iter"],
            "circuit_transpiled": circuit_transpiled
        }

        return reverse_config_trans

    def fit_transform(self, data: np.ndarray) -> np.ndarray:
        """
        Takes the data points and applies the PIT.

        :param data: Data samples
        :return: Transformed data points
        """
        df = pd.DataFrame(data)
        epit = df.copy(deep=True).transpose()
        self.reverse_epit_lookup = epit.copy(deep=True)

        epit.values[::] = [self.emp_integral_trans(row) for row in epit.values]
        epit = epit.transpose()
        self.reverse_epit_lookup.values[::] = [np.sort(row) for row in self.reverse_epit_lookup.values]

        df = epit.copy()
        self.reverse_epit_lookup = self.reverse_epit_lookup.values
        return df.values

    def _reverse_emp_integral_trans_single(self, values: np.ndarray) -> list[float]:
        """
        Takes one data point and applies the inverse PIT.

        :param values: Data point
        :return: Data point after applying the inverse transformation
        """
        values = values * (np.shape(self.reverse_epit_lookup)[1] - 1)
        rows = np.shape(self.reverse_epit_lookup)[0]

        # TODO THIS IS A TEMPORARY BUG FIX FOR RARELY OCCURRING CASES THAT LEAD TO INDEX_ERRORS.
        #  WILL BE FIXED IN RELEASE 2.1.3.
        # Ensure values are within bounds
        values_l = np.clip(np.floor(values).astype(int), 0, self.reverse_epit_lookup.shape[1] - 1)
        values_h = np.clip(np.ceil(values).astype(int), 0, self.reverse_epit_lookup.shape[1] - 1)

        # if we are an integer then floor and ceiling are the same
        is_int_mask = 1 - (values_h - values_l)
        row_indexer = np.arange(rows)
        result_l = self.reverse_epit_lookup[
            ([row_indexer], [values_l])]  # doing 2d lookup as [[index1.row, index2.row],[index1.column, index2.column]]
        result_h = self.reverse_epit_lookup[
            ([row_indexer], [values_h])]  # where 2d index tuple would be (index1.row, index1.column)
        # lookup int or do linear interpolation
        return result_l * (is_int_mask + values - values_l) + result_h * (values_h - values)

    def inverse_transform(self, data: np.ndarray) -> np.ndarray:
        """
        Applies the inverse transformation to the full data set.

        :param data: Data set
        :return: Data set after applying the inverse transformation
        """
        res = [self._reverse_emp_integral_trans_single(row) for row in data]

        return np.array(res)[:, 0, :]

    def emp_integral_trans(self, data: np.ndarray) -> np.ndarray:
        """
        Applies the empirical integral transformation to the given data.

        :param data: Data points
        :return: Empirically transformed data points
        """
        rank = np.argsort(data).argsort()
        length = data.size
        ecdf = np.linspace(0, 1, length, dtype=np.float64)
        ecdf_biject = ecdf[rank]
        return ecdf_biject



import logging

from quark.BenchmarkManager import Instruction
from quark.modules.Core import Core
from quark.modules.applications.Application import Application


class InstructionDemo(Application):
    """
    A simple QUARK Application implementation showing the usage of instructions.
    """

    def __init__(self, application_name: str = None):
        super().__init__(application_name)
        self.submodule_options = ["Dummy"]

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple:
        """
        Preprocess input data with given configuration and instructions.

        :param input_data: Data to be  processed.
        :param config: Configuration for processing the data
        :param kwargs: Additional keyword arguments
        :return: Instruction, processed data, and processing time.
        """
        logging.info("%s", kwargs.keys())
        logging.info("previous_job_info: %s", kwargs.get("previous_job_info"))

        rep_count = kwargs["rep_count"]
        instruction_name = config.get("instruction", Instruction.PROCEED.name)
        instruction = Instruction.PROCEED

        if instruction_name == Instruction.PROCEED.name:
            instruction = Instruction.PROCEED
        elif instruction_name == Instruction.INTERRUPT.name:
            instruction = Instruction.INTERRUPT
        if instruction_name == "mixed":
            instruction = Instruction.PROCEED
            if rep_count % 2 == 1:
                instruction = Instruction.INTERRUPT
        elif instruction_name == "exception":
            raise Exception("demo exception")

        logging.info(
            "InstructionDemo iteration %s returns instruction %s",
            rep_count, instruction.name
        )
        return instruction, "", 0.

    def get_parameter_options(self) -> dict:
        """
        Returns parameter options for the preprocess method.
        """
        return {
            "instruction": {
                "values": [
                    Instruction.PROCEED.name,
                    Instruction.INTERRUPT.name,
                    "exception",
                    "mixed"
                ],
                "description": "How should preprocess behave?"
            }
        }

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule for the given option.

        :param option: The submodule option
        :return: Default submodule
        """
        return Dummy()

    def save(self, path: str, iter_count: int) -> None:
        """
        Saves the current state to the specified path.

        :param path: Path where the state should be saved
        :param iter_count: Iteration count.
        """
        pass


class Dummy(Core):
    """
    Dummy QUARK module implementation which is used by the InstructionDemo.
    """

    def get_parameter_options(self) -> dict:
        """
        Returns parameter options for the Dummy module.

        :return: Dictionary containing parameter options
        """
        return {}

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule for the given option.

        :param option: The submodule option
        """
        pass



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations  # Needed if you want to type hint a method with the type of the enclosing class

import os
import sys
from abc import ABC, abstractmethod
from typing import final

from quark.utils import _get_instance_with_sub_options
from quark.Metrics import Metrics


class Core(ABC):
    """
    Core Module for QUARK, used by all other Modules that are part of a benchmark process.
    """

    def __init__(self, name: str = None):
        """
        Constructor method.

        :param name: Name used to identify this QUARK module. If not specified class name will be used as default.
        """
        self.submodule_options = []
        self.sub_options = []
        self.depending_parameters = False
        self.preprocessed_input = None
        self.postprocessed_input = None
        if name is None:
            name = self.__class__.__name__
        self.name = name
        self.metrics = Metrics(name, os.path.relpath(sys.modules[self.__module__].__file__))

    @abstractmethod
    def get_parameter_options(self) -> dict:
        """
        Returns the parameters for a given module.

        Should always be in this format:

        .. code-block:: json

            {
               "parameter_name":{
                  "values":[1, 2, 3],
                  "description":"How many nodes do you need?"
               },
                "parameter_name_2":{
                  "values":["x", "y"],
                  "description":"Which type of problem do you want?"
               }
            }

        :return: Available settings for this application
        """
        raise NotImplementedError("Please don't use the base version of get_parameter_options. "
                                  "Implement your own override instead.")

    def get_available_submodules(self, option: list) -> list:
        """
        If the module has submodules depending on certain options, this method should adjust the submodule_options
        accordingly.

        :param option: List of chosen options
        :return: List of available submodules
        """
        return []

    def get_depending_parameters(self, option: str, config: dict) -> dict:
        """
        If the module has parameters depending on certain options, this method should return the parameters for the
        given option.

        :param option: The chosen option
        :param config: Current config dictionary
        :return: The parameters for the given option
        """
        return {}

    @final
    def get_submodule(self, option: str) -> Core:
        """
        Submodule is instantiated according to the information given in self.sub_options.
        If self.sub_options is None, get_default_submodule is called as a fallback.

        :param option: String with the options
        :return: Instance of a module
        """
        if self.sub_options is None or not self.sub_options:
            return self.get_default_submodule(option)
        return _get_instance_with_sub_options(self.sub_options, option)

    # TODO Think if the naming of get_default_submodule can be improved to better reflect its function.
    @abstractmethod
    def get_default_submodule(self, option: str) -> Core:
        """
        Given an option string by the user, this returns a submodule.

        :param option: String with the chosen submodule
        :return: Module of type Core
        """
        raise NotImplementedError("Please don't use the base version of get_default_submodule. "
                                  "Implement your own override instead.")

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Essential method for the benchmarking process. This is always executed before traversing down
        to the next module, passing the data returned by this function.

        :param input_data: Data for the module, comes from the parent module if that exists
        :param config: Config for the module
        :param kwargs: Optional keyword arguments
        :return: The output of the preprocessing and the time it took to preprocess
        """
        return input_data, 0.0

    def postprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Essential Method for the benchmarking process. Is always executed after the submodule is finished. The data by
        this method is passed up to the parent module.

        :param input_data: Input data comes from the submodule if that exists
        :param config: Config for the module
        :param kwargs: Optional keyword arguments
        :return: The output of the postprocessing and the time it took to postprocess
        """
        return input_data, 0.0

    @final
    def get_available_submodule_options(self) -> list:
        """
        Gets the list of available options.

        :return: List of module options
        """
        if self.sub_options is None or not self.sub_options:
            return self.submodule_options
        else:
            return [option["name"] for option in self.sub_options]

    @staticmethod
    def get_requirements() -> list:
        """
        Returns the required pip packages for this module. Optionally, version requirements can be added.

        :return: List of dictionaries
        """
        return []
